/*
 * MEU.java
 * Top-level application launcher.
 * Created on July 29, 2000, 2:36 AM
 * This is free open source software distributed under the IBM Public License found
 * on the World Wide Web at http://oss.software.ibm.com/developerworks/opensource/license10.html
 * Copyright *C* 2000, Jack J. Woehr, PO Box 51, Golden, CO 80402-0051 USA jax@well.com
 * Copyright *C* 2000, International Business Machines Corporation and others. All Rights Reserved.
 */

package com.SoftWoehr.JTOpenContrib.MEU;

// import the connection manager
import com.SoftWoehr.JTOpenContrib.connectionmgr.*;


/** Create an MEU editing session.
 * @author jax
 * @version 1.0
 */
public class MEU extends Object {

  /** Identify version of release */
  public static final String MEU_VERSION = "1.0";

  /** Connection manager for AS400 connections */
  private Mgr connectionManager = new Mgr();

  /** Collection of frames we manage */
  private java.util.Vector meuFrames = new java.util.Vector(4,4);

  /** Properties used by editor */
  private MEUProperties properties;

  /** Keep track of the name of the property file we loaded at startup */
  private String propertiesFileLoaded;

  /** Hold our QCDemo ops nav window */
  private com.SoftWoehr.JTOpenContrib.QCDemo.QCPartnerFrame qcDemo = null;

  private boolean allowedToExit = true;

  /** Creates new MEU */
  public MEU() {
    /* So that we are ready to answer properties queries
     * before anything else gets started
     */
    properties = new MEUProperties();
  }

  /** Creates new MEU on a properties file, server name, ifs path. Any
   * of these values may legitimately be null.
   * @param propertiesFileName Relative or full pathname of MEU properties file to open.
   * @param serverName Name of server to access.
   * @param ifsPath Fully qualified IFS pathname of file member to edit.
   */
  public MEU(String propertiesFileName, String serverName, String ifsPath) {
    this();
    if (null != propertiesFileName) {
      loadProperties(propertiesFileName);
    }
    /** Open the first window */
    defaultSession(serverName, ifsPath);
  }

  /** Creates a new MEU with the desired properties
   * @param propertiesFileName MEU properties file name to open and parse and use.
   */
  public MEU (String propertiesFileName) {
    properties = new MEUProperties(propertiesFileName);
  }

  /** Fetch a server for an MEUEditBuffer ... the ServiceRecord object contains
   * a ref to an AS400 object.
   * @param systemName System one needs an AS400 object for.
   * @param client The instance which needs the service record (AS400 obj).
   * @throws UnknownServiceException If the service requested is unknown.
   * @throws UnknownProtocolException If the protocol requested is unknown.
   * @return A ServiceRecord which contains the AS400 object.
   */
  public ServiceRecord getServer(String systemName, ServiceClient client)
  throws Mgr.UnknownServiceException, Mgr.UnknownProtocolException
  {
    return connectionManager.getService("mgr:" + systemName + "/RECORDACCESS", client);
  }

  /** Let go of a server for an MEUEditBuffer
   * @param serviceRecord The original return from getServer()
   * @throws InvalidServiceRecordException If the Mgr doesn't know anything about the sr fed to it.
   */
  public void freeServer(ServiceRecord serviceRecord)
  throws Mgr.InvalidServiceRecordException
  {
    connectionManager.freeService(serviceRecord);
  }

  /** Open a first empty edit window */
  public void defaultSession() {
    try {
      newFrame
      (new java.awt.Point
      (getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.open.xlocation"),
      getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.open.ylocation")));
    }
    catch (MEUProperties.NoSuchPropertyException e) {
      e.printStackTrace(System.err);
      newFrame(null);
    }
    catch (MEUProperties.InvalidIntPropertyException e) {
      e.printStackTrace(System.err);
      newFrame(null);
    }
  }

  /** Open a first edit window on a server and file member
   * @param serverName The server on which the member resides.
   * @param ifsPath The fully-qualified source physical file member IFS path.
   */
  public void defaultSession(String serverName, String ifsPath) {
    try {
      newFrame
      (new java.awt.Point
      (getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.open.xlocation"),
      getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.open.ylocation")),
      serverName,
      ifsPath);
    }
    catch (MEUProperties.NoSuchPropertyException e) {
      e.printStackTrace(System.err);
      newFrame(null);
    }
    catch (MEUProperties.InvalidIntPropertyException e) {
      e.printStackTrace(System.err);
      newFrame(null);
    }
  }

  /** Add an edit frame to the collection to be closed */
  private void addFrame(MEUFrame m) {
    meuFrames.add(m);
  }

  /** Remove an edit frame from the collection we watch */
  private void removeFrame (MEUFrame m) {
    if (meuFrames.contains(m)) {
      meuFrames.remove(m);
      System.gc();
    }
    else {
      System.err.println("Non-existent frame requested for MEU removal.");
    }
  }

  /** Create a new frame, add it to the collection MEU manages, and show it
   * @param location Position on desktop for window to appear.
   */
  public void newFrame(java.awt.Point location) {
    MEUFrame m = new MEUFrame(this);
    addFrame (m);
    if (null != location) {
      m.setLocation(location);
    }
    m.show();
  }

  /** Create a new frame, add it to the collection MEU manages, open a file member and show it
   * @param location Position on desktop for window to appear. May be null, then default position used.
   * @param serverName Name of server to open on. May be null, in which case default or none is used.
   * @param ifsPath The fully-qualified source physical file member IFS path.
   */
  public void newFrame(java.awt.Point location, String serverName, String ifsPath) {
    MEUFrame m = new MEUFrame(this, serverName, ifsPath);
    addFrame (m);
    if (null != location) {
      m.setLocation(location);
    }
    m.show();
  }

  /** Load the properties from a props file.
   * @param filename Java-style properties file to load from.
   */
  public void loadProperties(String filename) {
    try {
      properties.loadFromFile(filename);
      propertiesFileLoaded = filename;
    }
    catch (java.io.FileNotFoundException e) {
      MEUStockDialog.showExceptionMessageDialog(null, "Couldn't load properties from " + filename, e);
    }
    catch (java.io.IOException e) {
      MEUStockDialog.showExceptionMessageDialog(null, "Couldn't load properties from " + filename, e);
    }
  }

  /** Store the properties
   * @param filename Write current user-set properties to a file.
   */
  public void storeProperties(String filename) {
    if (null == filename) {
      filename = propertiesFileLoaded;
    }
    if (null != filename) {
      try {
        properties.storeToFile(filename);
      }
      catch (java.io.FileNotFoundException e) {
        MEUStockDialog.showExceptionMessageDialog(null, "Couldn't store properties to " + filename, e);
      }
      catch (java.io.IOException e) {
        MEUStockDialog.showExceptionMessageDialog(null, "Couldn't store properties to " + filename, e);
      }
    }
    else {
      MEUStockDialog.showErrorMessageDialog
      (null, "No properties file exists yet, choose Save As from Settings menu", "Problem saving settings");
    }
  }
  /** Get a property stored by this application controller
   * @param property Name of property
   * @return String value or null.
   */
  public String getProperty(String property) {
    return properties.getProperty(property);
  }

  /** Set a property stored by this application controller.
   * Return the previous value for this property.
   * @param property Name of property to set
   * @param value Value of property to set
   * @return Previous value of property.
   */
  public String setProperty(String property, String value) {
    String previous = getProperty(property);
    properties.setProperty(property, value);
    return previous;
  }

  /** Get a specific int property.
   * @param property The name of the property
   * @throws NoSuchPropertyException There ain't no such
   * @throws InvalidIntPropertyException The property value string isn't convertible to a base 10 integer
   * @return The integer value of the property.
   */
  public int getIntProperty (String property)
  throws MEUProperties.NoSuchPropertyException,
  MEUProperties.InvalidIntPropertyException
  {
    return properties.getIntProperty(property);
  }

  /** Cascade all frames
   * @return Next point to which a window would be cascaded after all
   * have been cascaded.
   */
  public java.awt.Point cascade() {

    // Defaults in case property throws.
    java.awt.Point nextLocation = new java.awt.Point(0,0);
    int xInc = 10;
    int yInc = 10;

    // Get the properties.
    try {
      nextLocation = new java.awt.Point
      (getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.open.xlocation"),
      getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.open.ylocation"));
      xInc = getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.cascade.xincrement");
      yInc = getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.cascade.yincrement");
    }
    catch (MEUProperties.NoSuchPropertyException e) {
      e.printStackTrace(System.err);
    }
    catch (MEUProperties.InvalidIntPropertyException e) {
      e.printStackTrace(System.err);
    }

    // Cascade the windows.
    for (java.util.Enumeration e = meuFrames.elements(); e.hasMoreElements(); ) {
      MEUFrame f = (MEUFrame) e.nextElement();
      f.setLocation(nextLocation);
      f.toFront();
      nextLocation.translate(xInc, yInc);
    }
    return nextLocation;
  }

  /** Instance QCDemo if necessary and show window
   * @return The QCDemo instance shared by all windows.
   */
  public com.SoftWoehr.JTOpenContrib.QCDemo.QCPartnerFrame getQCDemo() {
    if (null == qcDemo) {
      qcDemo = new com.SoftWoehr.JTOpenContrib.QCDemo.QCPartnerFrame (connectionManager);
      qcDemo.setLocation(40,40);
    }
    return qcDemo;
  }

  /** Bring up an operations navigator
   * @param evt The actionEvent triggering the bringup.
   */
  public void eventQCDemo(java.awt.event.ActionEvent evt) {
    getQCDemo().setVisible(true);
    getQCDemo().setState(javax.swing.JFrame.NORMAL);
    getQCDemo().toFront();
  }

  /** If set false, closes which would do System.exit() don't.
   * @param tf true if closes may System.exit() if appropriate; false if not.
   */
  public void setAllowedToExit (boolean tf) {
    allowedToExit = tf;
  }

  /** If set false, closes which would do System.exit() don't.
   * @return true if MEU is allowed to do System.exit() when it would
   * be appropriate in a standalone setting.
   * @see #setAllowedToExit
   */
  public boolean isAllowedToExit () {
    return allowedToExit;
  }

  /** Exit if allowed */
  private void exitIfAllowed(int retcode) {
    if (allowedToExit) {
      System.exit(retcode);
    }
  }

  /** An MEU Frame has signalled its MEUController that it wants to exit.
   * The MEUController has cleaned up and now requests to close. The
   * frame must be removed from our collection.
   * @param f Frame making request
   * @return true if close may proceed.
   */
  public boolean closeRequested (MEUFrame f) {
    removeFrame(f);
    if (0 == meuFrames.size()) {
      exitIfAllowed (0);
    }
    return true;
  }

  /** This is the end, my friend! */
  private void performSystemExit() {
    for (java.util.Enumeration e = meuFrames.elements(); e.hasMoreElements(); ) {
      MEUFrame m = (MEUFrame) e.nextElement();
      m.cleanup();
      m.dispose();
    }
    meuFrames.removeAllElements();
    System.gc();
    exitIfAllowed(0);
  }

  /** Iterate through frames requesting shutdown. If all approve
   * (true), i.e., if the user doesn't CANCEL on a "Save?" requester,
   * go ahead and have each frame cleanup, dispose each frame,
   * delete the heck out of them as well as one can, all this
   * to make sure connection resources etc. are returned. Then
   * exitIfAllowed();
   * @param f Frame making request.
   * @return True if shutdown proceeding
   * @see #isAllowedToExit
   * @see #setAllowedToExit
   */
  public boolean requestShutdown(MEUFrame f) {
    boolean result = true;
    for (java.util.Enumeration e = meuFrames.elements(); e.hasMoreElements(); ) {
      result &= (((MEUFrame) e.nextElement()).requestShutdown(f));
      if (!result) {
        break;
      }
    }
    if (result) {
      performSystemExit();
    }
    return result;
  }

  /********/
  /* MAIN */
  /********/

  /** Runs the SoftWoehr MEU editor
   * @param argv The command line for the editor.
   */
  public static void main(String[] argv) {
    String propertiesFileName = null;
    String serverName = null;
    String ifsPath    = null;

    /* Set properties file name, if any */
    for (int i = 0; i < argv.length; i++) {
      if (argv[i].substring(0,2).equals("-p")) {
        propertiesFileName = argv[i].substring(2);
        break;
      }
    }
    /* Set server name to use, if any */
    for (int i = 0; i < argv.length; i++) {
      if (argv[i].substring(0,2).equals("-s")) {
        serverName = argv[i].substring(2);
        break;
      }
    }
    /* Set ifsPath to use, if any */
    for (int i = 0; i < argv.length; i++) {
      if (!argv[i].substring(0,1).equals("-")) {
        ifsPath = argv[i];
        break;
      }
    }

    /** Open on */
    MEU meu = new MEU(propertiesFileName, serverName, ifsPath);
  }
}
