/*
 * MEUCommand.java
 * Represents commands entered on the command line.
 * Created on September 1, 2000, 9:21 PM
 * This is free open source software distributed under the IBM Public License found
 * on the World Wide Web at http://oss.software.ibm.com/developerworks/opensource/license10.html
 * Copyright *C* 2000, Jack J. Woehr, PO Box 51, Golden, CO 80402-0051 USA jax@well.com
 * Copyright *C* 2000, International Business Machines Corporation and others. All Rights Reserved.
 */

package com.SoftWoehr.JTOpenContrib.MEU;

/** Embodies command parsing from the command line at
 * the top of the MEU editor.
 * @author jax
 * @version 1.0
 */
public class MEUCommand extends Object {

  /** The entire command string as passed in */
  private String commandString;

  /** The uppercased command */
  private String command;

  /** Holds the individual args to the command in original case */
  private java.util.Vector args = new java.util.Vector(3,3);

  ///////////////////////////////////////
  // Special processing for change specs
  // Only valid immediately after a
  // successful processChangeSpecs()
  ///////////////////////////////////////

  /** If processChangeSpecs() is called successfully, the change separator char */
  private String changeSeparator;

  /** If processChangeSpecs() is called successfully, the change target (sought string to change) */
  private String changeTarget;

  /** If processChangeSpecs() is called successfully, the change mod (string to change to) */
  private String changeMod;

  /** If processChangeSpecs() is called successfully, the  "*" or numerical times-to-change on line */
  private String changeLineRange;

  /** If processChangeSpecs() is called successfully, the  "*" or numerical times-to-change in buffer */
  private String changeBufferRange;

  /////////////////////////////////////////
  // Special processing for move/copy specs
  // Only valid immediately after a
  // successful processMoveCopySpecs()
  /////////////////////////////////////////

  /** If processMoveCopySpecs() is called successfully, the first line to copy/move */
  private int moveCopyStart;

  /** If processMoveCopySpecs() is called successfully, the number of lines to copy/move */
  private int moveCopyNumber;

  /** If processMoveCopySpecs() is called successfully, the line move/copy before/after */
  private int moveCopyTarget;

  /** If processMoveCopySpecs() is called successfully, if true, move before, otherwise, after. */
  private boolean beforeNotAfter;

  /** Parse a string to create a new MEUCommand
   * @param s The command as entered by user.
   */
  public MEUCommand(String s) {
    commandString = s;
    java.util.StringTokenizer st = new java.util.StringTokenizer(s);
    if (st.hasMoreElements()) {
      command = ((String) st.nextElement()).toUpperCase();
    }
    while (st.hasMoreElements())
    {
      args.add(st.nextElement());
    }
  }

  /** Get the original command string
   * @return What the user entered.
   */
  public String getCommandString() {
    return commandString;
  }

  /** Get the command
   * @return Uppercased COMMAND from the command string entered by user.
   */
  public String getCommand () {
    return command;
  }

  /** Get the number of args to the command
   * @return number of space-delimited args.
   */
  public int numArgs() {
    return args.size();
  }

  /** Get the nth argument
   * @param n Zero-based index of argument (not including command itself).
   * @return Argument or null.
   */
  public String nthArg(int n) {
    return (String) args.get(n);
  }

  /** Get the arg string without the command
   * @return Arugment portion of user entry, as originally entered (no case conversion).
   */
  public String getArgString () {
    String s = commandString.trim();
    return s.substring(java.lang.Math.min(commandString.indexOf(' ') + 1, s.length()));
  }

  /** Used by a CHANGE command to process the change specification.
   * After this returns 'true' the getChange*() funcs are valid, before, not.
   * @return True on success.
   */
  public boolean processChangeSpecs() {
    boolean result = false;
    String spec = getArgString(); // Get rest of command string besides command ...
    if (null != spec) {
      if (!spec.equals("")) { // ... if any.
        changeSeparator = spec.substring(0,1); // Instance separator
        spec = spec.substring(1); // Rest of string past initial separator
        int nextSeparatorIndex = spec.indexOf(changeSeparator); // find next instance
        if (-1 != nextSeparatorIndex) { // If there is one ...
          changeTarget = spec.substring(0,nextSeparatorIndex); // ... instance the change target
          spec = spec.substring(++nextSeparatorIndex); // Bump past next separator
          if (spec != "") { // If we have any string left
            nextSeparatorIndex = spec.indexOf(changeSeparator);
            if (-1 != nextSeparatorIndex) { // If trailing separator found ...
              changeMod = spec.substring(0,nextSeparatorIndex);
              result = true; // Because now we have enough that we have a valid change spec
              spec = spec.substring(++nextSeparatorIndex); // Bump past trailing separator
              if (spec != "") { // If we have any string left
                java.util.StringTokenizer st = new java.util.StringTokenizer(spec);
                if (st.hasMoreTokens()) { // Look for ranges
                  changeLineRange = st.nextToken(); // Line ramge
                  if (st.hasMoreTokens()) {
                    changeBufferRange = st.nextToken(); // Buffer range
                  }
                }
              }
            }
          }
        }
      }
    }
    return result;
  }

  /** Return the change target from a change specification instanced by successful processChangeSpecs().
   * @return The string looked for to change.
   */
  public String getChangeTarget() {
    return changeTarget;
  }

  /** Return the change mod (the changed string) from a change specification instanced by successful processChangeSpecs().
   * @return The change text to replace the sought string.
   */
  public String getChangeMod() {
    return changeMod;
  }

  /** Return the "*" or number spec for number of changes in line instanced by successful processChangeSpecs().
   * @return "*" or number spec for number of changes in line
   */
  public String getChangeLineRange() {
    return changeLineRange;
  }

  /** Returns int number of changes or -1 for "*". If not a "*" and
   * non-convertible, throws.
   * @param changeRange String value for this parameter as entered by user.
   * @throws NumberFormatException <I>changRange</I> was neither "*" nor a base10 int.
   * @return Number of changes or -1 for "*"
   */
  public static int getConvertedChangeRange(String changeRange)
  throws NumberFormatException
  {
    int result = -1; // Meaning "any"
    if (!changeRange.equals("*")) { // Not "*", must be an int
      result = Integer.parseInt(changeRange);
    }
    return result;
  }

  /** Return the "*" or number spec for number of changes in buffer instanced by successful processChangeSpecs().
   * @return Number of rows to change or -1 for "*"
   */
  public String getChangeBufferRange () {
    return changeBufferRange;
  }

  /** Used by a MOVE or COPY command to process the specification.
   * After this returns 'true' the getMoveCopy*() funcs are valid, before, not.
   * @throws NumberFormatException If some numeric value in the specs is not a base10 int.
   * @return true on success.
   */
  public boolean processMoveCopySpecs()
  throws NumberFormatException
  {
    boolean result = false;
    // If we have enough args
    if (numArgs() == 4) {

      // Instance the numerical specs
      moveCopyStart  = Integer.parseInt(nthArg(0));
      moveCopyNumber = Integer.parseInt(nthArg(1));
      moveCopyTarget = Integer.parseInt(nthArg(3));

      // May have to rectify the before/after spec.
      String beforeAfter = nthArg(2).trim().toUpperCase();
      // Instance before/after spec
      beforeNotAfter = beforeAfter.equals("BEFORE");

      // If there was a legal before/after spec
      if (beforeNotAfter || beforeAfter.equals("AFTER")) {
        result = true;
      }
    }
    return result;
  }

  /** If processMoveCopySpecs() is called successfully, the first line to copy/move
   * @return First line number to move/copy.
   */
  public int getMoveCopyStart() {
    return moveCopyStart;
  }

  /** If processMoveCopySpecs() is called successfully, the number of lines to copy/move
   * @return number of lines to move/copy.
   */
  public int getMoveCopyNumber() {
    return moveCopyNumber;
  }

  /** If processMoveCopySpecs() is called successfully, the line move/copy before/after.
   * @return Number of line to move/copy before/after
   */
  public int getMoveCopyTarget() {
    return moveCopyTarget;
  }

  /** If processMoveCopySpecs() is called successfully, if true, move before, otherwise, after.
   * @return true if BEFORE
   * false if AFTER
   */
  public boolean getBeforeNotAfter() {
    return beforeNotAfter;
  }

  /** Show and test use of MEUCommand
   * @param argv Command line for MEU.
   */
  public static void main (String argv[]) {
    MEUCommand c = new MEUCommand(argv[0]);
    System.out.println("Command string is: " + c.getCommandString());
    System.out.println("Command is:        " + c.getCommand());
    System.out.println("Number of args is: " + c.numArgs());
    System.out.println("Arg string is    : " + c.getArgString());
    if (c.getCommand().equals("CHANGE")) {
      if (c.processChangeSpecs()) {
        System.out.println("Change target is : " + c.getChangeTarget());
        System.out.println("Change mod is    : " + c.getChangeMod());
        System.out.println("Line range is    : " + c.getChangeLineRange());
        System.out.println("Buffer range is  : " + c.getChangeBufferRange());
      }
      else {
        System.out.println("Invalid change specs");
      }
    }
    if (c.getCommand().equals("COPY")) {
      if (c.processMoveCopySpecs()) {
        System.out.println("Copy start is : " + c.getMoveCopyStart());
        System.out.println("Copy num is   : " + c.getMoveCopyNumber());
        System.out.println("Copy targ is  : " + c.getMoveCopyTarget());
        System.out.println("Before-Not-After is : " + c.getBeforeNotAfter());
      }
      else {
        System.out.println("Invalid copy specs");
      }
    }
  }
}
