/*
 * MEUController.java
 *
 * This is a frame controller.
 * There is one MEUController per frame.
 * There is also a top-level application
 * controller com.SoftWoehr.JTOpenContrib.MEU.MEU
 * Created on July 29, 2000, 2:35 AM
 * This is free open source software distributed under the IBM Public License found
 * on the World Wide Web at http://oss.software.ibm.com/developerworks/opensource/license10.html
 * Copyright *C* 2000, Jack J. Woehr, PO Box 51, Golden, CO 80402-0051 USA jax@well.com
 * Copyright *C* 2000, International Business Machines Corporation and others. All Rights Reserved.
 */

package com.SoftWoehr.JTOpenContrib.MEU;

import com.SoftWoehr.JTOpenContrib.connectionmgr.*;

/** Frame ontroller portion of MEU application; centralizes communications between
 * visual components and non-visual entities. There is also a top-level application
 * controller com.SoftWoehr.JTOpenContrib.MEU.MEU
 * @author jax
 * @version 1.0
 */
public class MEUController extends Object implements ServiceClient {

  /** Top-level application controller */
  private MEU meu;

  /** The service record which holds our AS400 */
  private ServiceRecord serviceRecord;

  /** The edit buffer we work on. */
  private MEUEditBuffer editBuffer;

  /** The frame to which the controller pertains. */
  private MEUFrame frame;

  /************/
  /* Creation */
  /************/

  /** Creates new MEUController */
  public MEUController(MEU m) {
    setMEU (m);
  }

  /** Set the top-level application controller */
  private void setMEU (MEU m) {
    meu = m;
  }

  /** Get the top-level application controller */
  public MEU getMEU () {
    return meu;
  }

  /*********************/
  /* Server Management */
  /*********************/

  /** Get the AS400 provided us by the Mgr via the MEU application controller */
  private com.ibm.as400.access.AS400 getSystem () {
    if (null != serviceRecord) {
      return serviceRecord.as400;
    }
    else {
      return null;
    }
  }

  /** Get the name of the AS400 provided us by the Mgr via the MEU application controller */
  private String getSystemName () {
    if (null != serviceRecord) {
      return serviceRecord.serverName;
    }
    else {
      return null;
    }
  }

  /** Get an AS400 from the MEU application controller.
   * Don't bother re-requesting if we already have it.
   */
  private void setSystem(String systemName) {
    try {
      if (null != serviceRecord) {
        if (!serviceRecord.serverName.equals(systemName)) {
          freeSystem();
          serviceRecord = meu.getServer(systemName, this);
        }
      }
      else {
        serviceRecord = meu.getServer(systemName, this);
      }
    }
    catch (Mgr.UnknownServiceException e) {
      e.printStackTrace();
    }
    catch (Mgr.UnknownProtocolException e) {
      e.printStackTrace();
    }
  }

  /** Set the server (i.e., do setSystem) safely, dialoging where necessary. */
  public void setServer (String server) {
    if (null != server) {
      if (!server.trim().equals("")) {
        if (closeCancelEditSession ()) {
          setSystem(server);
          setFrameTitle();
          String dfltUser = meu.getProperty("com.SoftWoehr.JTOpenContrib.MEU.system.server.user");
          if (null != dfltUser) {
            getSystem().removeDefaultUser(getSystemName());
            getSystem().setDefaultUser(getSystemName(), dfltUser);
          }
          MEUStockDialog.showInfoMessageDialog (frame, "Server set to " + getSystemName(), "Server Changed");
        }
        else {
          MEUStockDialog.showInfoMessageDialog (frame, "You cancelled the operation.", "Server Not Changed");
        }
      }
      else {
        MEUStockDialog.showInfoMessageDialog (frame, "No server name entered.", "Server Not Changed");
      }
    }
    else {
      MEUStockDialog.showInfoMessageDialog (frame, "You cancelled the operation.", "Server Not Changed");
    }
  }

  /** Give back our instance of an AS400 */
  private void freeSystem() {
    if (null != serviceRecord) {
      try {
        meu.freeServer(serviceRecord);
      }
      catch (Mgr.InvalidServiceRecordException e) {
        e.printStackTrace();
      }
    }
  }

  /** Look up the default and set the server accordingly.
   * Frame must be set first because set server changes title.
   */
  public void setDefaultServer() {
    String dfltServerName = meu.getProperty("com.SoftWoehr.JTOpenContrib.MEU.system.server.servername");
    if (null != dfltServerName) {
      setServer(dfltServerName);
    }
  }

  /** Called by a Mgr, this function indicates that a service record supplied
   * earlier by the MGr to the ServiceClient instance will no longer be valid
   * after this call completes.
   * @param sr The service record which the Mgr is indicating will no longer be valid after the <B>relinquish</B> completes.
   */
  public void relinquish(ServiceRecord sr) {
    /** Probably other stuff should go here, shutting down files, etc */
    closeEditSession();
    serviceRecord = null;
  }

  /************************/
  /* Component Management */
  /************************/

  /** Get the frame to which this controller is attached */
  public MEUFrame getFrame () {
    return frame;
  }

  /** Set the frame we're associated with */
  public void setFrame (MEUFrame f) {
    frame = f;
  }

  /** Set title displayed to include file member */
  private void setFrameTitle (String lib, String file, String member) {
    frame.setTitle
    (getProductName() + " - " + getSystemName() + ":" + makeOS400Path(lib , file, member));
  }

  /** Set the title displayed when no file member. */
  private void setFrameTitle () {
    StringBuffer sb = new StringBuffer(getProductName());
    if (null != getSystemName()) {
      sb.append (" - " + getSystemName() + ":");
    }
    frame.setTitle (sb.toString());
  }

  private int getNumColumnsInView() {
    return editBuffer.getNumColumnsInView();
  }

  private int getNumRowsInView() {
    return editBuffer.getNumRowsInView();
  }

  /** Return the display row number
   * the cursor is on.
   */
  public int getDisplayedRowNumber() {
    return getTextArea().getDisplayedRowNumber();
  }

  /** Convert a display row number to an absolute row/record
   * number. -1 would equal "start of data" line and row ==
   * getNumberOfRecords === "end of data" line.
   */
  public int rowNumberToRecordNumber (int rownum) {
    return rownum + getTopRowDisplayed();
  }

  /** Return the record/row number
   * the cursor is on.
   * -1 means the start-of-data line
   */
  public int getCurrentRowNumber () {
    return rowNumberToRecordNumber(getDisplayedRowNumber());
  }

  /** Returns offset of cursor in srcdta on current line.
   * Floored at zero, max is srcdta length.
   */
  public int getOffsetInSRCDTA(int curpos) {
    return getTextArea().getOffsetInSRCDTA(curpos);
  }

  /** Move cursor on a line */
  public void setOffsetInSRCDTA(int offset) {
    getTextArea().setOffsetInSRCDTA(offset);
  }

  /** Get current text area cursor offset */
  public int getCurPos () {
    return getTextArea().getCurPos();
  }

  /** Set current cursort pos to wrapped cursor pos offered. */
  public void setCurPos(int pos) {
    getTextArea().setCurPos(getTextArea().wrappedCursorPos(pos));
  }

  /** Change view size area to match rows and columns
   * e.g., when loading a new file member.
   */
  private void adjustSize() {
    /* Debug */
    // System.out.println("Adjust with " + editBuffer.getNumColumnsInView() + " columns and " + editBuffer.getNumRowsInView() + " rows in view.");
    frame.resetSize(getNumColumnsInView(), getNumRowsInView());
  }

  /** Get the text component */
  private MEUTextArea getTextArea () {
    return frame.getTextArea();
  }

  /******************/
  /* The Status Bar */
  /******************/

  /** Get the status bar */
  private MEUStatusPanel getStatusBar() {
    return frame.getStatusBar();
  }

  /** Set message in message area of status bar */
  private void statusMessage(String message) {
    getStatusBar().statusMessage(message);
  }

  /** Set this visible status */
  private void statusAll() {
    getStatusBar().statusAll();
  }

  /** Set this visible status */
  private void statusAll(String message) {
    getStatusBar().statusAll(message);
  }

  /** Set this visible status */
  private void statusChangedIndicator() {
    getStatusBar().statusChangedIndicator();
  }

  /** Set this visible status */
  private void statusInsertButton(boolean f) {
    getStatusBar().statusInsertButton(f);
  }

  /******************************/
  /* Enabling, Insert and Focus */
  /******************************/

  /** Enable/disable text editing */
  public void setEditingEnabled(boolean enabled) {
    // System.out.println("Editing Enabled == " + getTextArea().isEnabled());
    getTextArea().setEnabled(enabled);
    // System.out.println("Editing Enabled == " + getTextArea().isEnabled());
  }

  /** Return insert mode (true) or typeover (false) */
  public boolean getInserting() {
    return getTextArea().getInserting();
  }

  /** Toggle the insert mode flag */
  public boolean toggleInserting () {
    return getTextArea().toggleInserting ();
  }

  /** Flip focus between command line and text area */
  public void toggleFocus() {
    frame.toggleFocus();
  }

  /** Transfer the focus for the text area */
  public void textAreaFocus() {
    getTextArea().requestFocus();
  }

  /****************************/
  /* Handling the edit buffer */
  /****************************/
  /** Get IFS Path of our file member */
  public String getIFSPath() {
    return editBuffer.getIFSPath();
  }

  /** See if file member needs save */
  public boolean isChanged() {
    boolean result = false;
    if (null != editBuffer) {
      result = editBuffer.isDirty();
    }
    return result;
  }

  /** Return total number of row records */
  public int getNumberOfRows() {
    return editBuffer.getNumberOfRows();
  }

  /** Return total number of row records */
  public int getNumberOfRowsDisplayed() {
    return editBuffer.getNumberOfRowsDisplayed();
  }

  /** Return record number of row displaying in top row.
   * -1 means the "Start of Data" message.
   * numrecords + 1 means "End of Data" message
   */
  private int getTopRowDisplayed() {
    return editBuffer.getTopRowDisplayed();
  }

  /** Return record field length for SRCDTA */
  public int getSRCDTAlength() {
    return editBuffer.getSRCDTAlength();
  }

  /** Return SRCDTA contents of a give row/recordnumber */
  public String getSRCDTAcontents(int rownum) {
    return editBuffer.getSRCDTAcontents(rownum);
  }

  /** Insert row below current row */
  public boolean insertRow () {
    boolean result = false;
    int pos = getCurPos();
    if (null != editBuffer) {
      int currow = getCurrentRowNumber();
      int lastRow = getNumRowsInView() - 1;
      boolean needToScroll = (lastRow == getDisplayedRowNumber());
      processChangedSavingCursor(); // Don't process line commands, this may *be* a line command!
      if (!(getNumberOfRows() == currow)) {
        result = editBuffer.insertEmptyRow(currow+1);
        if (result) {
          display();
          setCurPos(pos);
          getTextArea().moveCursorDown();
          if (needToScroll) {
            scrollAhead();
          }
        }
        else {
          MEUStockDialog.showErrorMessageDialog
          (frame, "MEU was unable to insert a row", "Problem inserting row");
        }
      }
      else {
        MEUStockDialog.showErrorMessageDialog
        (frame, "MEU can't insert a row here, move up one row", "Problem inserting row");
      }
    }
    return result;
  }

  /** Remove a specific row/record by its visual display row number */
  public boolean removeRow (int displayRowNum) {
    int recordnum = rowNumberToRecordNumber (displayRowNum);
    int curpos = getCurPos();
    processChangedSavingCursor();
    boolean result = false;
    if (( -1 == recordnum) || (getNumberOfRows() == recordnum)) {
      MEUStockDialog.showErrorMessageDialog
      (frame, "MEU can't remove that row because\nit's just a marker on the display, not a real row.", "Problem removing row");
    }
    else {
      result = editBuffer.removeRow(recordnum);
      if (result) {
        editBuffer.setDirty();
        display();
      }
      else {
        MEUStockDialog.showErrorMessageDialog
        (frame, "There was some problem removing row " + displayRowNum + " (record " + recordnum + ")", "Problem removing row");
      }
    }
    setCurPos(curpos);
    return result;
  }

  /** Remove the row at current cursor in the display */
  public boolean removeRow () {
    return removeRow(getDisplayedRowNumber());
  }

  /** Load all records in */
  private boolean load() {
    // Indicate wait.
    // java.awt.Cursor currentCursor = frame.getCursor();
    // frame.setCursor(new java.awt.Cursor(java.awt.Cursor.WAIT_CURSOR));
    // frame.repaint();
    // System.out.println("loading ...");
    return editBuffer.load();
    // The wait is over.
    // frame.setCursor(currentCursor);
  }

  /** Load in a specific file member and display it
   * Return true if process completed.
   */
  private boolean loadMember (String lib, String file, String member) {
    boolean result = false;
    com.ibm.as400.access.QSYSObjectPathName pathName =
    new com.ibm.as400.access.QSYSObjectPathName(lib, file, member, "MBR");
    if (null != getSystem()) {
      try {
        if (MEUFileMember.pathExists(getSystem(), pathName)) {
          newEditBuffer (lib, file, member);
          setFrameTitle(lib, file, member);
          if (load()) { // If file loaded okay.
            display();
            getStatusBar().statusInsertButton(getInserting());
            result = true;
          }
          else { // Couldn't load
            setFrameTitle();
            editBuffer = null;
            System.gc();
          }
        }
        else { // File member doesn't exist
          result = false;
          MEUStockDialog.showWarningMessageDialog
          (
          frame,
          "The file member does not exist. Use MEU's NEW command to create it.",
          pathName.getPath() + " does not exist"
          );
        }
      }
      catch (java.io.IOException e) {
        result = false;
        MEUStockDialog.showExceptionMessageDialog (frame, "Problem encountered checking if " + pathName.getPath() + " exists", e);
      }
    }
    else {
      MEUStockDialog.showErrorMessageDialog
      (frame, "You must first set the server name before opening a file member.", "Problem loading file member");
    }
    return result;
  }

  /** Create a new edit buffer. System must be instanced first */
  public void newEditBuffer (String lib, String file, String member) {
    editBuffer = new MEUEditBuffer(meu, getSystem(), lib, file, member);
    try {
      editBuffer.setTopRowDisplayed(meu.getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.edittext.toprowdisplayed"));
      editBuffer.setNumRowsInView  (meu.getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.edittext.rows"));
    }

    catch(MEUProperties.NoSuchPropertyException e) {
      e.printStackTrace();
    }
    catch(MEUProperties.InvalidIntPropertyException e) {
      e.printStackTrace();
    }
  }

  /** If changed, write back to the server. True on success.
   * Called from the menu Save selection and from doCommandSAVE
   */
  public boolean saveToServer() {
    return saveToServer(false);
  }

  /** Write back to the server if changed or forced
   * True on success. Force write if (boolean forceWrite).
   * Called from the menu Save selection and from doCommandSAVE
   */
  public boolean saveToServer(boolean forceWrite) {
    processChangedSavingCursor(); // See if we have changes
    boolean wasDirty = isChanged();
    boolean wasSaved = editBuffer.write(forceWrite); // Writes if changed or forced.
    if (wasDirty & !wasSaved) {
      MEUStockDialog.showErrorMessageDialog(frame, "MEU was unable to write the changed records to the server", "Problem Saving To Server");
    }
    return wasSaved;
  }

  /** Save records to a user-chosen file member.
   */
  public boolean saveAsToServer(com.ibm.as400.access.QSYSObjectPathName pathName) {
    boolean result = false;

    /* First, make sure everything is saved */
    processChanged();

    /* Create new buffer to represent alternative file */
    MEUEditBuffer bufferNew =
    new MEUEditBuffer (meu, getSystem(), pathName.getLibraryName(), pathName.getObjectName(), pathName.getMemberName());

    try { // Does file exist? If so, should we overwrite?
      if (MEUFileMember.pathExists(getSystem(), pathName)) {
        result =
        (javax.swing.JOptionPane.YES_OPTION == MEUStockDialog.showOverwriteFileMemberDialog(getSystem(), pathName.getPath(), frame));
      }

      else { // File member didn't yet exist.
        if // Should we create it?
        (result =
        (javax.swing.JOptionPane.YES_OPTION == MEUStockDialog.showCreateFileMemberDialog(getSystem(), pathName.getPath(), frame)))
        {
          result = bufferNew.createMember(MEUStockDialog.showFileDescriptionDialog(frame));
        }
      }
    }

    catch (java.io.IOException e) {
      MEUStockDialog.showExceptionMessageDialog (frame, "Problem encountered checking if " + pathName.getPath() + " exists", e);
      result = false;
    }

    /* If we got the member okay, set record format and replace rows */
    if (result) {
      bufferNew.setRecordFormat();
      result = bufferNew.replaceAllRowsWith(editBuffer);
    }

    /* If that's okay, write it out and redisplay. Caller will do status */
    if (result) {
      bufferNew.setTopRowDisplayed(editBuffer.getTopRowDisplayed());
      editBuffer.closeSession(false); // Le roi est mort ...
      editBuffer=bufferNew;           // ... vive le roi!
      editBuffer.write();
      setFrameTitle(editBuffer.getLibraryName(), editBuffer.getFileName(), editBuffer.getMemberName());
      display();
      System.gc();
    }
    return result;
  }

  /** Save records to a user-chosen file member
   * File member must already exist!
   * File must have same record format!
   */
  public boolean newOnServer(com.ibm.as400.access.QSYSObjectPathName pathName) {
    boolean result = false;
    com.ibm.as400.access.QSYSObjectPathName sourceFilePathName =
    new com.ibm.as400.access.QSYSObjectPathName (pathName.getLibraryName(), pathName.getObjectName(), "FILE");

    /* First, make sure everything is ready to be saved */
    if (null != editBuffer) {
      processChanged();
    }

    /** Give user opportunity to save current session */
    if (closeCancelEditSession()) {

      /* Create new buffer to represent alternative file */
      MEUEditBuffer bufferNew =
      new MEUEditBuffer (meu, getSystem(), pathName.getLibraryName(), pathName.getObjectName(), pathName.getMemberName());

      try { // Does file exist? If so, should we overwrite?
        if (MEUFileMember.pathExists(getSystem(), pathName)) {
          result =
          (javax.swing.JOptionPane.YES_OPTION == MEUStockDialog.showOverwriteFileMemberDialog(getSystem(), pathName.getPath(), frame));
        }

        else { // File member didn't yet exist.
          if (MEUFileMember.pathExists(getSystem(), sourceFilePathName)) {
            if // Should we create it?
            (result =
            (javax.swing.JOptionPane.YES_OPTION == MEUStockDialog.showCreateFileMemberDialog(getSystem(), pathName.getPath(), frame))
            )
            {
              result = bufferNew.createMember(MEUStockDialog.showFileDescriptionDialog(frame));
            }
          }
          else {
            result = false;
            MEUStockDialog.showWarningMessageDialog
            (
            frame,
            "The source physical file does not exist.\nUse the OS/400 CRTSRCPF command to create it.",
            sourceFilePathName.getPath() + " does not exist"
            );
          }
        }
      }

      catch (java.io.IOException e) {
        MEUStockDialog.showExceptionMessageDialog (frame, "Problem encountered checking if " + pathName.getPath() + " exists", e);
        result = false;
      }

      /* If we got the member okay, set record format and create a few empty rows */
      if (result) {
        bufferNew.setRecordFormat();
        result = bufferNew.appendEmptyRow();
      }

      /* If that's okay, write it out and redisplay. Caller will do status */
      if (result) {
        bufferNew.write();
        editBuffer=bufferNew;
        setFrameTitle(editBuffer.getLibraryName(), editBuffer.getFileName(), editBuffer.getMemberName());
        display();
        System.gc();
      }
    }
    return result;
  }

  /** Insert records from a user-chosen file member
   * File member must already exist!
   * File must have same record format!
   */
  public boolean getAllFrom(com.ibm.as400.access.QSYSObjectPathName pathName) {
    boolean result = false;

    /* First, make sure everything is saved */
    processChangedSavingCursor();

    /* Create new buffer to represent alternative file */
    MEUEditBuffer bufferNew =
    new MEUEditBuffer (meu, getSystem(), pathName.getLibraryName(), pathName.getObjectName(), pathName.getMemberName());

    try { // Does file exist? If so, should we overwrite?
      result = (MEUFileMember.pathExists(getSystem(), pathName));
    }

    catch (java.io.IOException e) {
      MEUStockDialog.showExceptionMessageDialog (frame, "Problem encountered checking if " + pathName.getPath() + " exists", e);
      result = false;
    }

    /* If we got the member okay, set record format and replace rows */
    if (result) {
      // In other words, if we're on the Start of Data line and below thereafter,
      // insert after. If we're on the last line, pretend we were on the previous
      // line and insert after.
      int rownum = java.lang.Math.min(getCurrentRowNumber() + 1, getNumberOfRows());
      bufferNew.load();
      result = editBuffer.insertAllRowsFrom(rownum, bufferNew);
      bufferNew.closeSession(false);
      editBuffer.setTopRowDisplayed(rownum);
    }

    else {
      MEUStockDialog.showErrorMessageDialog
      (frame,
      "Couldn't get file member " + pathName.getPath(),
      "Problem getting file member");
    }

    display();
    System.gc();
    return result;
  }

  /** Write the member out to a local file */
  public boolean writeLocaltext(java.io.File f) {
    boolean result = true;
    try {
      java.io.FileOutputStream fo = new java.io.FileOutputStream(f);
      editBuffer.writeLocaltext(fo);
      fo.flush();
      fo.close();
    }
    catch (java.io.FileNotFoundException e) {
      MEUStockDialog.showExceptionMessageDialog (frame, "Couldn't find file " + f.getName(), e);
      result = false;
    }
    catch (java.io.IOException e) {
      MEUStockDialog.showExceptionMessageDialog (frame, "Problem writing out " + f.getName(), e);
      result = false;
    }

    return result;
  }

  /*******************************/
  /* Opening and closing session */
  /*******************************/

  /** Open an edit session if the user doesn't cancel a save/discard of the previous contents */
  public boolean openEditSession(String ifsPath) {
    boolean result = false;
    try {
      com.ibm.as400.access.QSYSObjectPathName qPathName = new com.ibm.as400.access.QSYSObjectPathName(ifsPath);
      result = loadMember (qPathName.getLibraryName(), qPathName.getObjectName(), qPathName.getMemberName());
    }
    catch (com.ibm.as400.access.IllegalPathNameException e) {
      MEUStockDialog.showExceptionMessageDialog (frame, "This doesn't appear to be a valid pathname: " + ifsPath, e);
    }
    return result;
  }

  /** Close the edit buffer and release resources
   * checking with the user whether to save if necssary. */
  private void closeEditSession() {
    boolean writeOut = false;
    if (null != editBuffer) {
      if (isChanged()) {
        if (javax.swing.JOptionPane.YES_OPTION == MEUStockDialog.showCloseFileMemberDialog(getSystem(), getIFSPath(), frame)) {
          writeOut = true;
        }
      }
      editBuffer.closeSession(writeOut);
      editBuffer = null;
      System.gc();
    }
  }

  /** Close current session */
  private void closeSession (boolean saveNoSave) {
    editBuffer.closeSession(saveNoSave);
    editBuffer = null;
    setFrameTitle();
    display("");
    setEditingEnabled(false);
    System.gc();
  }

  /** Dialogs with the user if necessary
   * and allows user to save, discard or cancel a close.
   * Returns true after a save or discard, returns false for a cancel.
   */
  private boolean closeCancelEditSession () {
    boolean result = true;
    if (null != editBuffer) {
      if (isChanged()) {

        switch (MEUStockDialog.showCloseCancelFileMemberDialog (getSystem(), getIFSPath(), frame)) {

          case javax.swing.JOptionPane.YES_OPTION :
          closeSession(true);
          break;

          case javax.swing.JOptionPane.NO_OPTION :
          closeSession(false);
          break;

          case javax.swing.JOptionPane.CANCEL_OPTION :
          result = false;
          break;

          default :
          result = false;
        }
      }
      else {
        closeSession(false); // Buffer existed but may be discarded.
      }
    }
    return result;
  }


  /*****************************************************/
  /* Complex dialoging operations leading to record io */
  /*****************************************************/

  /** Offer to create a file member and do so if accepted */
  public void offerNewFileMember() {
    String path = null;
    if (null != getSystem()) {

      /* Dialog with user about file to open */
      path = MEUStockDialog.showOpenFileMemberDialog(
      getSystem(),
      meu.getProperty("com.SoftWoehr.JTOpenContrib.MEU.system.server.ifsdirectory"),
      frame
      );
      if (null != path) {
        if (newOnServer(new com.ibm.as400.access.QSYSObjectPathName(path))) {
          if (openEditSession (path)) {
            setEditingEnabled(true);
          }
          else {
            MEUStockDialog.showErrorMessageDialog
            (frame,
            "Couldn't open an editing session on " + path,
            "Problem loading file member");
          }
        }
      }
    }
    else {
      MEUStockDialog.showErrorMessageDialog
      (frame, "You must first set the server name before creating a file member.", "Problem creating file member");
    }
  }

  /** Offer to open a file member and do so if accepted */
  public void offerOpenFileMember() {
    String path = null;
    if (null != getSystem()) {

      /* Dialog with user about file to open */
      path = MEUStockDialog.showOpenFileMemberDialog(
      getSystem(),
      meu.getProperty("com.SoftWoehr.JTOpenContrib.MEU.system.server.ifsdirectory"),
      frame
      );

      // path = "/QSYS.LIB/JWOEHRS.LIB/MEUTEST.FILE/HELLO.MBR";
      if (null != path) {
        if (null != editBuffer) {
          processChanged(); // Do any updates necessary
          // Ignore line commands
        }
        if (closeCancelEditSession()) { // Offer user chance to save changed.
          if (openEditSession(path)) {
            // System.out.println("Setting Editing Enabled True in menuOpenFileMember()");
            setEditingEnabled(true);
            // textAreaFocus();
          }
          else {
            MEUStockDialog.showErrorMessageDialog
            (frame, "Couldn't open an editing session on " + path, "Problem loading file member");
          }
        }
        else {
          MEUStockDialog.showInfoMessageDialog
          (frame, "Because you cancelled the close,\n no new file member was loaded.", "Close cancelled");
        }
      }
    }
    else {
      MEUStockDialog.showErrorMessageDialog
      (frame, "You must first set the server name before opening a file member.", "Problem browsing for file member");
    }
  }

  /** Offer to Save As on server. Return true if successful. */
  protected boolean offerSaveAs() {
    boolean result = false;

    String path =
    MEUStockDialog.showSaveAsFileMemberDialog
    (
    getSystem(),
    meu.getProperty("com.SoftWoehr.JTOpenContrib.MEU.system.server.ifsdirectory"),
    frame
    );

    if (null != path) {
      com.ibm.as400.access.QSYSObjectPathName pathName = new com.ibm.as400.access.QSYSObjectPathName (path);
      result = saveAsToServer(pathName);
    }
    return result;
  }

  /** Offer to Save As on server. Return true if successful. */
  protected boolean offerGetAll() {
    boolean result = false;

    String path =
    MEUStockDialog.showOpenFileMemberDialog
    (
    getSystem(),
    meu.getProperty("com.SoftWoehr.JTOpenContrib.MEU.system.server.ifsdirectory"),
    frame
    );

    if (null != path) {
      com.ibm.as400.access.QSYSObjectPathName pathName = new com.ibm.as400.access.QSYSObjectPathName (path);
      result = getAllFrom(pathName);
    }
    return result;
  }


  /** Offer the user to save text format locally. Return true if completed. */
  public boolean offerSaveLocaltext() {
    boolean result = false;
    processChangedSavingCursor(); // Discard line commands
    java.io.File localfile = MEUStockDialog.showSaveLocalFileMemberDialog( ".", frame);
    if (null != localfile) {
      if (!localfile.isDirectory()) { // Not trying to save to a directory
        if (!localfile.isFile()) { // Not pre-existent file
          try { // Create previously nonexistent file .
            localfile.createNewFile();
            result = true;
          }
          catch (java.io.IOException e) { // Couldn't create file.
            MEUStockDialog.showExceptionMessageDialog (frame, "Couldn't create new file " + localfile.getPath(), e);
          }
        }
        else { // File exists -- does user want to overwrite?
          if (javax.swing.JOptionPane.OK_OPTION == MEUStockDialog.showOverwriteExtantLocalFileDialog(localfile.getPath(), frame)) {
            result = localfile.canWrite(); // Is it writable?
          }
        }
        if (result) { // If created or writable, write out.
          result = writeLocaltext(localfile); // need to report error
          if (!result) {
            MEUStockDialog.showErrorMessageDialog
            (frame, "An error was encountered when MEU attempting to write the file " + localfile.getPath() + ".", "MEU : Problem Writing File");
          }
        }
      }
      else { // User tried to save to directory, not file
        MEUStockDialog.showErrorMessageDialog
        (frame, "The path " + localfile.getPath() + " is not a file, it's a directory.", "MEU : Problem Writing File");
      }
      if (!result) {
        MEUStockDialog.showErrorMessageDialog
        (frame, "The file was not written.", "MEU : Problem Writing File");
      }
    }
    return result;
  }

  /** Offer to open a file member and do so if accepted */
  public boolean offerSearch() {
    boolean result = false;
    String searchString = MEUStockDialog.showSearchDialog(frame);
    if (null != searchString) {
      result = search(searchString);
    }
    return result;
  }

  /*********************************/
  /* Displaying a view on the text */
  /*********************************/

  /** Display the view */
  private void display() {
    if (null != editBuffer) {
      /** Show text */
      // getTextArea().setText(editBuffer.toDisplayString());
      adjustSize();
      replaceDisplay();
    }
  }

  /** Display some contents ... can be used for messages */
  private void display(String s) {
    getTextArea().setText(s);
  }

  /** Save cursor and do redisplay */
  public void clearRedisplaySavingCursor() {
    int pos = getCurPos();
    display();
    getTextArea().setCurPos(pos);
    getTextArea().wrapCursor();
  }

  /** Jump to a line */
  private void gotoLine(int seqnum) {
    seqnum = java.lang.Math.max(1, java.lang.Math.min(seqnum, getNumberOfRows()));
    int pos = getCurPos();
    editBuffer.setTopRowDisplayed(seqnum-1);
    display();
    setCurPos(pos);
  }

  /** Jump to top */
  private void gotoTop() {
    int pos = getTextArea().getCurPos();
    editBuffer.setTopRowDisplayed(-1);
    display();
    setCurPos(pos);
  }

  /** Jump to a line */
  private void gotoBottom() {
    int pos = getTextArea().getCurPos();
    editBuffer.setTopRowDisplayed(getNumberOfRows() - 1);
    display();
    setCurPos(pos);
  }

  /** Replace one row */
  private void replaceRow
  (int memberRow, int displayRow, boolean isProtected, boolean isChanged, String lineCommand) {
    String s = editBuffer.toString(memberRow);
    getTextArea().replaceVisibleRow(displayRow, s, isProtected, isChanged, lineCommand);
  }

  /** Replace all rows in display with this portion of edit */
  private void replaceDisplay() {
    int start = getTopRowDisplayed();
    int postindex =  java.lang.Math.min(start + getNumRowsInView(), getNumberOfRows() + 1);
    int nRows = postindex + 1 - (start + 1); // To avoid the -1 trap for top message row.
    // System.out.println("Rows == " + nRows );
    getTextArea().setToBlank(nRows);
    for (int i = 0; i < nRows; i++) {
      boolean isProtected = i+start == -1 | i+start == getNumberOfRows();
      // System.out.println("Row number == " + i + " and isProtected == " + isProtected);
      replaceRow // Replace row setting protected if it's the first or last
      ( i+start, i, isProtected, false, null);
    }
  }

  /*********************/
  /* Search & Replace */
  /*********************/

  /** Find offset of a string in a row. */
  private int searchOffsetInRow(String s, int startOffset, int rownum) {
    String contents = getSRCDTAcontents(rownum);
    int result = contents.substring(java.lang.Math.min(startOffset, getSRCDTAlength())).indexOf(s);
    if (-1 != result) {
      result += startOffset;
    }
    return result;
  }

  /** Jump to found string, if found */
  private boolean search (String s) {
    boolean result = false;
    processChangedSavingCursor(); // Discard line commands
    int rownum = java.lang.Math.max(0, getCurrentRowNumber());
    int foundRowNum = rownum;
    int curpos  = getCurPos();
    int offset  = getOffsetInSRCDTA(curpos + 1); // So we don't keep finding same string.
    // getOffsetInSRCDTA() floors and ceilings so it' s okay even if we're not in srcdta.
    int foundOffset = searchOffsetInRow(s, offset, rownum);
    if (-1 == foundOffset) {
      foundRowNum = editBuffer.search(s, rownum + 1);
      if (-1 != foundRowNum) {
        foundOffset = searchOffsetInRow(s, 0, foundRowNum);
      }
    }
    if (-1 != foundOffset) { // Did we ever manage to find it?
      gotoLine(foundRowNum + 1); // gotoLine is kinda funny since user puts in 1-based line number.
      // Setting curpos this way works because found line is now top of form.
      display(); // Gotta do this before we set cursor
      setCurPos(foundOffset + getTextArea().offset_to_srcdta);
      result = true;
    }
    else {
      display();
    }
    return result;
  }

  /** Subr for doCommandCHANGE() -- Change on instance on one row */
  private void changeOneOnCurrentRow(MEUCommand command, int currentRowNumber) {
    // Do the change and get new offset
    int newCursorOffset = editBuffer.changeNextOnLine
    (command.getChangeTarget(), command.getChangeMod(), getCurrentRowNumber(), getOffsetInSRCDTA(getCurPos()));

    // If offset indicates success move cursor, display and status it.
    if (-1 != newCursorOffset) {
      setOffsetInSRCDTA(newCursorOffset);
      int newcurpos = getCurPos();
      display();
      setCurPos(newcurpos);
      statusAll("1 changed");
    }
    else {
      statusAll("No changes");
    }
  }

  /** Subr for doCommandCHANGE() -- Change n instances on n rows */
  private void changeMultipleRows(MEUCommand command, int currentRowNumber) {
    int numChanged = 0; // number finally having been changed
    int lastRow = currentRowNumber; // Row we wish the cursor to appear on when we are done.
    int numToChange  = command.getConvertedChangeRange(command.getChangeLineRange()); // number we have to change per row
    int numberOfRows = command.getConvertedChangeRange(command.getChangeBufferRange()); // number of rows to change
    int offset = getOffsetInSRCDTA(getCurPos());  // Offset in srcdta. Start at current position.
    int lastOffset = offset; // Offset at which we wish cursor to appear in row.
    String target = command.getChangeTarget(); // String sought
    String mod    = command.getChangeMod(); // String  to change to.

    // Rectify change specs
  if (-1 == numToChange)  { numToChange  = getSRCDTAlength(); }
  if (-1 == numberOfRows) { numberOfRows = getNumberOfRows(); }

    // Calc what will be the end row (one past the last one to do).
    int endRow = java.lang.Math.min(getNumberOfRows(), java.lang.Math.max(numberOfRows + currentRowNumber, 0));

    // Loop on rows
    for ( ; currentRowNumber < endRow; currentRowNumber++) {
      for (int i = 0; i < numToChange; i++) {
        offset= editBuffer.changeNextOnLine (target, mod, currentRowNumber, offset);
        if (-1 != offset) { // Found in this line
          numChanged++;
          lastRow = currentRowNumber;
          lastOffset = offset;
        }
        else { // Not found in this line
          break;
        }
      } // End loop on row
      offset = 0;
    } // End loop on rows.

    // If count indicates success move cursor, display and status it.
    if (0 < numChanged) {
      gotoLine(lastRow + 1); // This is a display line
      display();
      setCurPos(0); // Move cursor to top row, i.e., last row changed.
      setOffsetInSRCDTA(lastOffset); // Move cursor to correct position
      statusAll(numChanged + " changed");
    }
    else {
      statusAll("No changes");
    }
  }

  /** Subr for doCommandCHANGE() -- Change n instances on one row */
  private void changeMultipleOnCurrentRow(MEUCommand command, int currentRowNumber)
  throws NumberFormatException
  {
    int cursorOffset = -1;
    int numChanged = 0;
    int numToChange = command.getConvertedChangeRange(command.getChangeLineRange());
    if (-1 == numToChange) { // If "*" spec, change as many as found
      numToChange = getSRCDTAlength(); // effectively, max possible to change, we'll break if no change.
    }
    // Change while more to change if any to change.
    for (; numToChange > 0; numToChange--) {
      // Do the change and get new offset
      cursorOffset= editBuffer.changeNextOnLine
      (command.getChangeTarget(), command.getChangeMod(), getCurrentRowNumber(), getOffsetInSRCDTA(getCurPos()));
      if (-1 != cursorOffset) {
        numChanged++;
        setOffsetInSRCDTA(cursorOffset);
      }
      else {
        break;
      }
    }
    // If count indicates success move cursor, display and status it.
    if (0 < numChanged) {
      int newcurpos = getCurPos();
      display();
      setCurPos(newcurpos);
      statusAll(numChanged + " changed");
    }
    else {
      statusAll("No changes");
    }
  }

  /**********************/
  /* Moving and Copying */
  /**********************/

  /** Copy moveCopyNumber lines starting at moveCopyStart before/after moveCopy Target */
  private boolean copyRange (int moveCopyStart, int moveCopyNumber, int moveCopyTarget, boolean beforeNotAfter) {
    boolean result = false;
    int i;
    if (0 <= moveCopyStart && moveCopyStart < getNumberOfRows()) {
      if (0 < moveCopyNumber && moveCopyNumber <= getNumberOfRows() - moveCopyStart) {
        if (0 <= moveCopyTarget && moveCopyTarget < getNumberOfRows()) {
          if (beforeNotAfter && moveCopyTarget > 0) { // Copy before
            for (i = 0; i < moveCopyNumber; i++) {
              if (!editBuffer.copyRowBefore(moveCopyStart + i, moveCopyTarget + i))
              {
                MEUStockDialog.showErrorMessageDialog
                (frame,
                "line " + (moveCopyStart + i) + " before line " + (moveCopyTarget + i),
                "Problem copying rows");
                break;
              }
              result = true;
            }
          } // End copy before
          // Now we have to test to make sure the preceding didn't fail due to before + row 0.
          else if (!beforeNotAfter) { // Copy after
            for (i = 0; i < moveCopyNumber; i++) {
              if (!editBuffer.copyRowAfter(moveCopyStart + i, moveCopyTarget + i))
              {
                MEUStockDialog.showErrorMessageDialog
                (frame,
                "line " + (moveCopyStart + i) + " after line " + (moveCopyTarget + i),
                "Problem copying rows");
                break;
              }
              result = true;
            }
          } // End copy after
        } // End legal target
      } // End legal number
    } // End legal start
    return result;
  }

  /** Move moveCopyNumber lines starting at moveCopyStart before/after moveCopy Target */
  private boolean moveRange (int moveCopyStart, int moveCopyNumber, int moveCopyTarget, boolean beforeNotAfter) {
    boolean result = false;
    int i;
    if (0 <= moveCopyStart && moveCopyStart < getNumberOfRows()) {
      if (0 < moveCopyNumber && moveCopyNumber <= getNumberOfRows() - moveCopyStart) {
        if (0 <= moveCopyTarget && moveCopyTarget < getNumberOfRows()) {
          if (beforeNotAfter && moveCopyTarget > 0) { // Move before
            for (i = 0; i < moveCopyNumber; i++) {
              if (!editBuffer.moveRowBefore
              (
              moveCopyStart < moveCopyTarget ? moveCopyStart  : moveCopyStart  + i,
              moveCopyStart < moveCopyTarget ? moveCopyTarget : moveCopyTarget + i
              ))
              {
                MEUStockDialog.showErrorMessageDialog
                (frame,
                "line " + (moveCopyStart + i) + " before line " + (moveCopyTarget + i),
                "Problem moving rows");
                break;
              }
              result = true;
            }
          } // End move before
          // Now we have to test to make sure the preceding didn't fail due to before + row 0.
          else if (!beforeNotAfter) { // Copy after
            for (i = 0; i < moveCopyNumber; i++) {
              if (!editBuffer.moveRowAfter
              (
              moveCopyStart < moveCopyTarget ? moveCopyStart  : moveCopyStart  + i,
              moveCopyStart < moveCopyTarget ? moveCopyTarget : moveCopyTarget + i
              ))
              {
                MEUStockDialog.showErrorMessageDialog
                (frame,
                "line " + (moveCopyStart + i) + " after line " + (moveCopyTarget + i),
                "Problem moving rows");
                break;
              }
              result = true;
            }
          } // End move after
        } // End legal target
      } // End legal number
    } // End legal start

    return result;
  }

  /***************************/
  /* Line Command Processing */
  /***************************/

  /** Process individual line command */
  private boolean processLineCommand(String cmd, int rownum) {

    /* Most line commands end all line commands */
    boolean continueProcessing = true;

    /* Sometimes there's a message */
    String message = null;

    /* Get the command */
    MEULineCommand lc = new MEULineCommand(cmd);

    // /* Debug */
    // /* Debug */ System.out.println ("Row " + rownum + " had line command: " + lc.command + " with numeric arg " + lc.numericArg);
    // /* Debug */

    // Used for loops
    int common_index = (null == lc.numericArg ? 1 : lc.numericArg.intValue());

    /* If there's no command, there might be a goto line number */
    if (null == lc.command) {
      if (null != lc.numericArg) {
        continueProcessing = false;
        gotoLine(lc.numericArg.intValue());
      }
    }

    /////////////////////////////////
    // Otherise, there was a command.
    /////////////////////////////////

    /* Insert */
    else if (lc.command.equals("I")) {
      for (int i = 0; i < common_index; i++) {
        continueProcessing = false;
        if (!insertRow()) {
          break;
        }
      }
    }
    /* Delete */
    else if (lc.command.equals("D")) {
      continueProcessing = false;
      for (int i = 0; i < common_index; i++) {
        if (!removeRow()) {
          break;
        }
      }
    }
    /* Move top */
    else if (lc.command.equals("T")) {
      continueProcessing = false;
      gotoTop();
    }
    /* Move bottom */
    else if (lc.command.equals("B")) {
      continueProcessing = false;
      gotoBottom();
    }
    /* Unknown command */
    else {
      continueProcessing = false;
      message = cmd + "?";
    }

    statusAll(message);
    return continueProcessing;
  }

  /** Process all line commands in the vector of all rows. */
  private void processLineCommands (java.util.Vector v) {
    int i = 0; // The vector has all the line commands by row.
    for (java.util.Enumeration e = v.elements() ; e.hasMoreElements() ;) {
      String cmd = ((String) e.nextElement()).trim();
      // System.out.println("Row == " + i + " Cmd == " + cmd);
      if (!cmd.equals("")) {
        /* Process command ... if command returns false,
         * done processing line commands
         */
        if (!processLineCommand(cmd, i)) {
          break;
        }
      }
      i++;
    }
  }

  /******************************************/
  /* Support for the display which requires */
  /* the attention of the controller.       */
  /******************************************/

  /* Get changes back from one row of the display */
  private void processChanged(int rownum) {
    int recordnum = rownum + getTopRowDisplayed();
    editBuffer.updateRow (recordnum, getTextArea().getSRCDTAcontents(rownum));
    getTextArea().setRowChanged(false, rownum);
  }

  /* Get changes back from all rows of the display */
  private void processChanged() {
    for (int i = 0; i < getNumberOfRowsDisplayed(); i++) {
      if (!getTextArea().getRowWriteProtected(i)) {
        if (getTextArea().getRowChanged(i)) {
          processChanged(i);
        }
      }
    }
    display();
  }

  /** Save cursor and process changes */
  public void processChangedSavingCursor() {
    int pos = getCurPos();
    processChanged();
    setCurPos(pos);
  }

  /** Process both changed and line commands */
  private void processAll() {
    /* Get line commands first because process changed replaces them */
    java.util.Vector v = getTextArea().getRowLineCommands();

    /* Process textually changed lines */
    processChangedSavingCursor();

    /* Do status */
    statusAll();

    /* Process line commands
     * May statusMessage();
     */
    processLineCommands(v);
  }

  /** Scroll ahead and display
   * CALLER MUST INSURE EDIT BUFFER IS VALID
   */
  private void scrollAhead() {
    int pos = getCurPos();
    editBuffer.scrollTopOfViewAhead();
    display();
    setCurPos(pos);
  }

  /** Scroll back and display
   * CALLER MUST INSURE EDIT BUFFER IS VALID
   */
  private void scrollBack() {
    int pos = getCurPos();
    editBuffer.scrollTopOfViewBack();
    display();
    setCurPos(pos);
  }

  /** Split or join a line at the cursor */
  private void splitJoin()  {
    if (null != editBuffer) {
      int curpos  = getCurPos();
      int offset  = getOffsetInSRCDTA(curpos);
      int currow  = getCurrentRowNumber();
      int lastrow = getNumberOfRows() - 1;
      if (0 <= currow && currow <= lastrow)  {
        editBuffer.splitJoinRow(currow, offset, lastrow);
      }
      display();
      setCurPos(curpos);
    }
  }

  /*******************/
  /* Menu Selections */
  /*******************/

  /////////////////////////////////////////////
  // The "offer..()" functions dialog the user.
  /////////////////////////////////////////////

  /**  Open file open via dialog in response to a menu item */
  public void menuNewFileMember(java.awt.event.ActionEvent evt) {
    offerNewFileMember();
    statusAll();
  }

  /**  Open file open via dialog in response to a menu item */
  public void menuOpenFileMember(java.awt.event.ActionEvent evt) {
    offerOpenFileMember();
    statusAll();
  }

  /** Set server via a dialog in response to a menu item */
  public void menuSetServer(java.awt.event.ActionEvent evt) {
    String s = MEUStockDialog.showSetServerDialog(frame);
    setServer(s);
    statusAll();
  }

  /** Process changed and save changes to server file member */
  public void menuSave (java.awt.event.ActionEvent evt) {
    if (null != editBuffer) {
      saveToServer();
    }
    statusAll();
  }

  /** Process changed and save under new name by dialog. */
  public void menuSaveAs (java.awt.event.ActionEvent evt) {
    if (null != editBuffer) {
      offerSaveAs();
    }
    statusAll();
  }

  /** Get a file member into the current file member by dialog. */
  public void menuGetAll(java.awt.event.ActionEvent evt) {
    if (null != editBuffer) {
      offerGetAll();
    }
    statusAll();
  }

  /** Save the current buffer as a local text file by dialog. */
  public void menuSaveLocal(java.awt.event.ActionEvent evt) {
    if (null != editBuffer) {
      offerSaveLocaltext();
    }
    statusAll();
  }

  /** Insert a new row after the current position. */
  public void menuInsertRow (java.awt.event.ActionEvent evt) {
    if (null != editBuffer) {
      insertRow();
    }
    else {
      beep();
    }
    statusAll();
  }

  /** Delete row at the current position. */
  public void menuRemoveRow (java.awt.event.ActionEvent evt) {
    if (null != editBuffer) {
      removeRow();
    }
    else {
      beep();
    }
    statusAll();
  }

  /** Search for a string. */
  public void menuSearch (java.awt.event.ActionEvent evt) {
    if (null != editBuffer) {
      if (offerSearch()) {
        statusAll("Found");
      }
      else {
        statusAll("Not found");
      }
    }
  }

  /***************/
  /* Key Events */
  /**************/

  /** Clear changed and line commands and redisplay */
  public void eventClearRedisplay(java.awt.event.KeyEvent e) {
    clearRedisplaySavingCursor();
    statusAll();
  }

  /** Process all changed rows */
  public void eventProcessChanged(java.awt.event.KeyEvent e) {
    processAll(); // Does a statusAll() and possibly statusMessage()
    System.gc(); // Lots of stuff created during processAll()!
  }

  /************/
  /* Actions */
  /************/

  /** Flip focus between command line and text area */
  public void actionToggleFocus(java.awt.event.ActionEvent e) {
    frame.toggleFocus();
  }

  /** Scroll ahead and display */
  public void actionScrollAhead(java.awt.event.ActionEvent e) {
    if (null != editBuffer) {
      processAll(); // Does a statusAll() and possibly statusMessage()
      scrollAhead();
    }
  }

  /** Scroll ahead and display */
  public void actionScrollBack(java.awt.event.ActionEvent e) {
    if (null != editBuffer) {
      processAll(); // Does a statusAll() and possibly statusMessage()
      scrollBack();
    }
  }

  /** Split/Join a line */
  public void actionSplitJoin(java.awt.event.ActionEvent e) {
    if (null != editBuffer) {
      processChangedSavingCursor(); // Discard line commands
      splitJoin(); // Does a display()
      statusAll();
    }
  }

  /** Clear changed and line commands and redisplay */
  public void actionToggleInserting(java.awt.event.ActionEvent e) {
    statusInsertButton(toggleInserting());
  }

  /** Save current file member */
  public void actionSave(java.awt.event.ActionEvent e) {
    if (null != editBuffer) {
      saveToServer();
      statusAll();
    }
  }

  /** Close the current frame (warning on save changed) */
  public void actionClose(java.awt.event.ActionEvent e) {
    iWantToClose();
  }

  /************************/
  /* Command interpreting */
  /************************/

  /** Execute a command issued at the editor command line */
  public void doCommand (String commandString) {
    MEUCommand command = new MEUCommand(commandString);
    String cmd = command.getCommand();

    if (null != cmd) {
      if      (cmd.equals("BOTTOM")) {
        doCommandBOTTOM (command);
      }
      else if (cmd.equals("CLOSE")) {
        doCommandCLOSE (command);
      }
      else if (cmd.equals("CHANGE")) {
        doCommandCHANGE (command);
      }
      else if (cmd.equals("COPY")) {
        doCommandCOPYMOVE (command);
      }
      else if (cmd.equals("DELETE")) {
        doCommandDELETE (command);
      }
      else if (cmd.equals("EDIT")) {
        doCommandLOAD (command);
      }
      else if (cmd.equals("FILE")) {
        doCommandFILE (command);
      }
      else if (cmd.equals("FONTSIZE")) {
        doCommandFONTSIZE (command);
      }
      else if (cmd.equals("GET")) {
        doCommandGET (command);
      }
      else if (cmd.equals("INSERT")) {
        doCommandINSERT (command);
      }
      else if (cmd.equals("JOIN")) {
        doCommandSPLITJOIN (command);
      }
      else if (cmd.equals("LOAD")) {
        doCommandLOAD (command);
      }
      else if (cmd.equals("MOVE")) {
        doCommandCOPYMOVE (command);
      }
      else if (cmd.equals("NEW")) {
        doCommandNEW (command);
      }
      else if (cmd.equals("SAVE")) {
        doCommandSAVE (command);
      }
      else if (cmd.equals("FIND")) {
        doCommandFIND (command);
      }
      else if (cmd.equals("SERVER") || cmd.equals("SYSTEM")) {
        doCommandSERVER (command);
      }
      else if (cmd.equals("SPLIT")) {
        doCommandSPLITJOIN (command);
      }
      else if (cmd.equals("TOP")) {
        doCommandTOP (command);
      }
      else { // Try to convert to a number and go to that line.
        try {
          int lineNumber = Integer.valueOf(cmd).intValue();
          gotoLine(lineNumber);
        }
        catch (NumberFormatException e) {
          MEUStockDialog.showCommandErrorDialog(frame, "Unknown command " + cmd, command.getCommandString());
        }
      }
      System.gc(); // If a command was issued, clean up afterwards.
    }
  }

  /** Perform the FONTSIZE command */
  private void doCommandFONTSIZE (MEUCommand command) {
    int fontsize = 0;
    if (command.numArgs() > 0) {
      try {
        fontsize = Integer.parseInt(command.nthArg(0));
        if (fontsize > 1 & fontsize < 73) {
          frame.resetFont(new java.awt.Font("Monospaced", java.awt.Font.PLAIN, fontsize));
        }
        else {
          MEUStockDialog.showCommandErrorDialog(frame, "FONTSIZE must be between 2 and 72.", command.getCommandString());
        }
      }
      catch (NumberFormatException e) {
        MEUStockDialog.showCommandErrorDialog(frame, "FONTSIZE command issued with invalid numeric argument.", command.getCommandString());
        // e.printStackTrace(System.err); // not needed
      }
    }
    else {
      MEUStockDialog.showCommandInfoMessageDialog
      (frame, "Font size is currently " + Integer.toString(frame.queryFontSize()) + ".", command.getCommandString());
    }
    statusAll();
  }

  /** Perform the LOAD command */
  private void doCommandLOAD (MEUCommand command) {
    if (null != getSystem()) {
      if (null != editBuffer) {
        processChanged();
      }
      // closeCancelEditSession() checks for edit buffer
      if (closeCancelEditSession()) { // Offer user chance to save changed.
        if (command.numArgs() == 3) {
          String lib    = command.nthArg(0).toUpperCase();
          String file   = command.nthArg(1).toUpperCase();
          String member = command.nthArg(2).toUpperCase();
          if (loadMember (lib, file, member)) {
            // System.out.println("Setting Editing Enabled True in doCommandLoad()");
            setEditingEnabled(true);
            // textAreaFocus();
          }
          else {
            MEUStockDialog.showErrorMessageDialog
            (frame,
            "Couldn't open an editing session on " + makeOS400Path(lib, file, member),
            "Problem loading file member");
          }
        }
        else {
          MEUStockDialog.showCommandErrorDialog(frame, "LOAD command syntax: LOAD lib file member", command.getCommandString());
        }
      }
      else {
        MEUStockDialog.showInfoMessageDialog
        (frame, "Because you cancelled the close,\n no new file member was loaded.", "Close cancelled");
      }
    }
    else {
      MEUStockDialog.showErrorMessageDialog
      (frame, "You must first set the server name before opening a file member.", "Problem loading file member");
    }
    statusAll();
  }

  /** Perform the SERVER command */
  private void doCommandSERVER(MEUCommand command) {
    if (command.numArgs() > 0) {
      setServer(command.nthArg(0));
    }

    else {
      MEUStockDialog.showCommandInfoMessageDialog
      (frame, "Server name is " + getSystemName() +  ".", command.getCommandString());
    }
    statusAll();
  }

  /** Perform the INSERT command */
  private void doCommandINSERT(MEUCommand command) {
    if (command.numArgs() == 1) {
      try {
        int num = Integer.parseInt(command.nthArg(0));
        for (int i = 0; i < num; i++) {
          insertRow();
        }
      }
      catch (NumberFormatException e) {
        e.printStackTrace();
      }
    }
    else if (command.numArgs() == 0) {
      insertRow();
    }
    else {
      MEUStockDialog.showCommandErrorDialog(frame, "INSERT command syntax: INSERT <nrows>", command.getCommandString());
    }
    statusAll();
  }

  /** Perform the DELETE command */
  private void doCommandDELETE(MEUCommand command) {
    if (command.numArgs() == 1) {
      try {
        int num = Integer.parseInt(command.nthArg(0));
        for (int i = 0; i < num; i++) {
          if (!removeRow()) {
            // Break if we had a problem (ran out of rows);
            break;
          }
        }
      }
      catch (NumberFormatException e) {
        e.printStackTrace();
      }
    }
    else if (command.numArgs() == 0) {
      removeRow();
    }
    else {
      MEUStockDialog.showCommandErrorDialog(frame, "DELETE command syntax: DELETE <nrows>", command.getCommandString());
    }
    statusAll();
  }

  /** Perform the TOP command */
  private void doCommandTOP(MEUCommand command) {
    gotoTop();
    statusAll();
  }

  /** Perform the BOTTOM command */
  private void doCommandBOTTOM(MEUCommand command) {
    gotoBottom();
    statusAll();
  }

  /** Perform the SAVE command */
  private void doCommandSAVE (MEUCommand command) {
    boolean was_saved = false;
    switch (command.numArgs()) {
      case 0 :
      was_saved = saveToServer();
      break;

      case 3 :
      was_saved = saveAsToServer
      (new com.ibm.as400.access.QSYSObjectPathName(command.nthArg(0), command.nthArg(1), command.nthArg(2), "MBR"));
      break;

      default :
      MEUStockDialog.showCommandErrorDialog(frame, "SAVE command syntax: SAVE <library file member>", command.getCommandString());
    }
    if (!was_saved) {
      MEUStockDialog.showErrorMessageDialog(frame, "The file was not saved.", "Problem saving file");
    }
    statusAll();
  }

  /** Save file member and close */
  private void doCommandFILE (MEUCommand command) {
    boolean was_saved = false;

    switch (command.numArgs()) {
      case 0 :
      was_saved = saveToServer(true);
      break;

      case 3 :
      was_saved = saveAsToServer
      (new com.ibm.as400.access.QSYSObjectPathName(command.nthArg(0), command.nthArg(1), command.nthArg(2), "MBR"));
      break;

      default :
      MEUStockDialog.showCommandErrorDialog(frame, "FILE command syntax: FILE <library file member>", command.getCommandString());
    }

    if (was_saved) {
      iWantToClose();
    }
    else {
      MEUStockDialog.showErrorMessageDialog(frame, "File not saved.", "FILE command cancelled");
    }
    statusAll();
  }

  /** Close ... warns on save of changed */
  private void doCommandCLOSE(MEUCommand command) {
    iWantToClose();
    statusAll();
  }

  /** Perform the LOAD command */
  private void doCommandNEW (MEUCommand command) {
    if (null != getSystem()) {
      if (command.numArgs() == 3) {
        String lib    = command.nthArg(0).toUpperCase();
        String file   = command.nthArg(1).toUpperCase();
        String member = command.nthArg(2).toUpperCase();

        // newOnServer() does the process changed and close.
        if (newOnServer(new com.ibm.as400.access.QSYSObjectPathName(lib, file, member, "MBR"))) {
          if (loadMember (lib, file, member)) {
            // System.out.println("Setting Editing Enabled True in doCommandLoad()");
            setEditingEnabled(true);
            // textAreaFocus();
          }
          else {
            MEUStockDialog.showErrorMessageDialog
            (frame,
            "Couldn't open an editing session on " + makeOS400Path(lib, file, member),
            "Problem loading file member");
          }
        }
        else {
          MEUStockDialog.showErrorMessageDialog
          (frame, "No new file member was created.", "Problem creating file member");
        }
      }
      else {
        MEUStockDialog.showCommandErrorDialog(frame, "NEW command syntax: NEW lib file member", command.getCommandString());
      }
    }
    else {
      MEUStockDialog.showErrorMessageDialog
      (frame, "You must first set the server name before creating a new file member.", "Problem creating file member");
    }
    statusAll();
  }

  /** Perform the LOAD command */
  private void doCommandGET (MEUCommand command) {
    if (null != getSystem() && (null != editBuffer)) {
      if (command.numArgs() == 3) {
        String lib    = command.nthArg(0).toUpperCase();
        String file   = command.nthArg(1).toUpperCase();
        String member = command.nthArg(2).toUpperCase();
        if (!getAllFrom(new com.ibm.as400.access.QSYSObjectPathName(lib, file, member, "MBR"))) {
          MEUStockDialog.showErrorMessageDialog(frame, "The file was not gotten.", "Problem getting file");
        }
      }
      else {
        MEUStockDialog.showCommandErrorDialog(frame, "GET command syntax: GET library file member", command.getCommandString());
      }
    }
    else {
      MEUStockDialog.showErrorMessageDialog
      (frame, "You must first be editing a file before GETting file member.", "Problem GETting file member");
    }
    statusAll();
  }

  /** Split or join a line */
  private void doCommandSPLITJOIN (MEUCommand command) {
    if (null != editBuffer) {
      processChangedSavingCursor(); // Discard line commands
      splitJoin(); // Does a display()
    }
    statusAll();
  }

  /** Search for next occurrence of a string anywhere in the buffer */
  private void doCommandFIND(MEUCommand command) {
    if (null != editBuffer) {
      if (command.numArgs() < 1) {
        MEUStockDialog.showCommandErrorDialog(frame, "FIND command syntax: FIND string", command.getCommandString());
      }
      else {
        if (search(command.getArgString())) {
          statusAll("Found");
        }
        else {
          statusAll("Not found");
        }
      }
    }
  }

  /** Search for next occurrence of a string anywhere in the buffer */
  private void doCommandCHANGE(MEUCommand command) {
    if (null != editBuffer) {
      processChangedSavingCursor(); // discard line commands
      boolean validChangeSpecs = command.processChangeSpecs();
      if (!validChangeSpecs) {
        MEUStockDialog.showCommandErrorDialog
        (frame, "CHANGE command syntax: CHANGE /old/new/ [ [times-per-line | *] [times-per-buffer | *] ]", command.getCommandString());
      }
      else { // Valid change  specs
        int currentRowNumber = getCurrentRowNumber();
        if (0 <= currentRowNumber && currentRowNumber < getNumberOfRows()) { // If the row is writable
          if (null == command.getChangeLineRange()) { // We have a non-ranged change spec
            changeOneOnCurrentRow(command, currentRowNumber);
          } // End change one time.
          else if (null == command.getChangeBufferRange()){ // We have line range but no buffer range
            try { // Throws if spec for number of changes isn't a number or *.
              changeMultipleOnCurrentRow(command, currentRowNumber);
            }
            catch (NumberFormatException e) { // Oops, user error.
              MEUStockDialog.showExceptionMessageDialog (frame, "Invalid times-per-line change spec\n(Must be number or *)\n" +
              command.getChangeLineRange(), e);
            }
          } // End multiple on one line.
          else { // We have line range and buffer range both
            try {// Throws if spec for number of changes isn't a number or *.
              changeMultipleRows(command, currentRowNumber);
            }
            catch (NumberFormatException e) { // Oops, user error.
              MEUStockDialog.showExceptionMessageDialog
              (frame,
              "Invalid lines-per-buffer or times-per-buffer change spec\n (Must be number or *):\n"
              + command.getChangeLineRange()
              + " " + command.getChangeBufferRange(),
              e);
            }
          } // End multiple line change
        } // End writable row offered.
        else { // Row not writable.
          MEUStockDialog.showCommandErrorDialog (frame, "You can't CHANGE anything on this line.", command.getCommandString());
        } // End not writable
      } // End valid change specs
    } // End valid edit buffer
  }

  /** Copy or move rows -- handles both commands */
  private void doCommandCOPYMOVE(MEUCommand command) {
    if (null != editBuffer) {
      processChangedSavingCursor(); // discard line commands
      try {
        if (command.processMoveCopySpecs()) {
          boolean result;
          if (command.getCommand().equals("COPY")) { // It's a COPY
            result =
            copyRange  // Start and targ are -1 since user sees ones-based line numbers.
            (command.getMoveCopyStart() - 1, command.getMoveCopyNumber(), command.getMoveCopyTarget() - 1, command.getBeforeNotAfter());
          }
          else { // It's a MOVE
            result =
            moveRange // Start and targ are -1 since user sees ones-based line numbers.
            (command.getMoveCopyStart() - 1, command.getMoveCopyNumber(), command.getMoveCopyTarget() - 1, command.getBeforeNotAfter());
          }
          if (result) // If MOVE or COPY okay
          {
            statusAll(command.getMoveCopyNumber() + " rows");
          }
          else { // Not okay
            statusAll("error");
          }
          clearRedisplaySavingCursor();
        }
        else { // Bad args -- effectively, bad BEFORE/AFTER spec
          MEUStockDialog.showCommandErrorDialog (frame, "COPY syntax: COPY from numrows BEFORE|AFTER row.", command.getCommandString());
        }
      }
      catch (NumberFormatException e) { // Bad numerical args
        MEUStockDialog.showExceptionMessageDialog (frame, "Invalid COPY numerical spec\n(E.g., COPY 1 2 AFTER 5)\n" +
        command.getCommandString(), e);
      }
    } // End has-edit-buffer
  }

  /*************/
  /* Utilities */
  /*************/

  /** Make the string representing an OS/400 LIB/FILE.MBR path */
  private String makeOS400Path (String lib, String file, String member) {
    return (lib + "/" + file + "." + member).toUpperCase();
  }

  /** Make the string representing the product name and level */
  private String getProductName () {
    return "SoftWoehr MyEditUtil " + MEU.MEU_VERSION;
  }

  /** Beep */
  public void beep() {
    getTextArea().beep();
  }

  /***********************/
  /* Closing and Exiting */
  /***********************/

  /** Close, warning for save of changed */
  protected void iWantToClose() {
    frame.iWantToClose();
  }

  /** An MEUFrame calls this when user tries to close frame.
   * MEUController checks if anything needs saving (or if user
   * decides to cancel). Then if all is still go, forwards close
   * request to application controller (class MEU).
   */
  public boolean closeRequested(MEUFrame f) {
    boolean result = true; // Default is "yes, close"
    if (null != getSystem()) {
      if (null != editBuffer) {
        processChangedSavingCursor(); // Do any updates necessary
        // Ignore line commands
        if (closeCancelEditSession()) { // Offer user chance to save changed.
          result = meu.closeRequested(f); // If no, function will return false.
        }
        else {
          result = false;
        }
      }
      else {
        // Server but no edit buffer, might as well ask to close.
        result = meu.closeRequested(f); // If no, function will return false.
      }
    }
    else { // No server, no edit buffer, might as well ask to close.
      result = meu.closeRequested(f); // If no, function will return false.
    }
    statusChangedIndicator();
    return result;
  }

  /** Iterate through frames requesting shutdown. If all approve
   * (true), i.e., if the user doesn't CANCEL on a "Save?" requester,
   * go ahead and have each frame cleanup, dispose each frame,
   * delete the heck out of them as well as one can, all this
   * to make sure connection resources etc. are returned. Then
   * System.exit(some result);
   */
  public boolean requestShutdown(MEUFrame f) {
    boolean result = true; // Default is "yes, close"
    // Debug
    // System.out.println("Shutdown requested controller " + this.toString() + " by frame " + f.toString());
    if (null != getSystem()) {
      if (null != editBuffer) {
        processChangedSavingCursor(); // Do any updates necessary
        // Ignore line commands
        result = closeCancelEditSession(); // Offer user chance to save changed.
      }
    }
    statusChangedIndicator();
    return result;
  }

  /** Clean up connection stuff */
  public void cleanup() {
    /* Relinquish connection */
    freeSystem();
    /* Release other AS400 stuff*/
    /* Release edit buffer */

    // Debug
    // System.out.println("Cleanup requested controller " + this.toString());
  }
}
