/*
 * MEUEditBuffer.java
 * This is the edit buffer that manipulates session records.
 * Created on July 29, 2000, 1:42 PM
 * This is free open source software distributed under the IBM Public License found
 * on the World Wide Web at http://oss.software.ibm.com/developerworks/opensource/license10.html
 * Copyright *C* 2000, Jack J. Woehr, PO Box 51, Golden, CO 80402-0051 USA jax@well.com
 * Copyright *C* 2000, International Business Machines Corporation and others. All Rights Reserved.
 */

package com.SoftWoehr.JTOpenContrib.MEU;

/** Handles row editing the rows of unicode text representing the content
 * of the file member being edited.
 * @author jax
 * @version 1.0
 */
public class MEUEditBuffer extends Object {

  /** Collection of editable rows. */
  private final MEUEditRow editRows;

  /** Reference to application controller */
  private final MEU meu;

  /** True if a row has been inserted/moved/altered/deleted  */
  private boolean dirty = false;

  /** What's the top row in the current view?
   * The value assigned here is a rough default.
   * The controller should call setTopRowDisplayed()
   */
  private int topRowDisplayed = -1 ;

  /** How many rows may be displayed in the current view?
   * The value assigned here is a rough default.
   * The controller should call setNumRowsInView()
   */
  private int numRowsInView = 25;

  /** The server our editable entity resides on. */
  private com.ibm.as400.access.AS400 as400;

  /** The identity of the edited entity */
  private String library;
  private String file;
  private String member;

  /** The AS400File entity which is being edited */
  private final MEUFileMember fileMember;

  /** Creates new MEUEditBuffer
   * @param a400 The AS400 object representing the server the file to be edited resides on.
   * @param rowlen The length of the data (text) portion of the row.
   * @param lib IFS name of the library in which the file member resides.
   * @param fil IFS name of the file in which the member resides.
   * @param mem Member to be edited.
   */
  public MEUEditBuffer(MEU anMEU, com.ibm.as400.access.AS400 a400, String lib, String fil, String mbr) {
    meu = anMEU;
    as400 = a400;
    library = lib;
    file = fil;
    member = mbr;
    fileMember = new MEUFileMember(as400, library, file, member);
    editRows = new MEUEditRow(fileMember);
  }

  /*****************/
  /* Member Access */
  /*****************/

  /** Get library name */
  public String getLibraryName() {
    return library;
  }

  /** Get file name */
  public String getFileName() {
    return file;
  }

  /** Get member name */
  public String getMemberName() {
    return member;
  }

  /** Return the edit rows associated with this edit buffer */
  protected MEUEditRow getEditRows() {
    return editRows;
  }

  /** A row has been inserted/moved/altered/deleted  */
  protected void setDirty() {
    dirty=true;
  }

  /** No row has been inserted/moved/altered/deleted  */
  protected void resetDirty() {
    dirty=false;
  }

  /** Is there a dirty row in the buffer? */
  public boolean isDirty() {
    return dirty;
  }

  /** Return the IFSPath of the file member we represent */
  public String getIFSPath  () {
    return editRows.getIFSPath();
  }

  /****************/
  /* Row handling */
  /****************/

  /** Get number of row records */
  public int getNumberOfRows() {
    return editRows.getNumberOfRows();
  }

  /** Return number of rows in display */
  public int getNumberOfRowsDisplayed() {
    // There are two extra rows in the display, indexed -1 and numrows respectively.
    // These extras are the start of data and end of data message lines
    return java.lang.Math.min((getNumberOfRows() + 2) - (getTopRowDisplayed() + 1), numRowsInView);
  }


  /** Return a string representation of all rows
   * @return The java string representation of the entire file member.
   */
  public String toString() {
    return toString(0, getNumberOfRows());
  }

  /** Return a string representation of a certain number of rows,
   *  starting zero-based at a certain row. The postindex is zero-based
   * and is the index of the next row after the last converted.
   *  This representation is suitable for MEUTextArea.setText(),
   *  that is, it has the extra fields and has linefeeds embedded at
   * the end of all rows but the last.
   * @param startRow The number of the row to start with. This is zero-based.
   * @param numRows How many rows to format as a string. If 0 is the
   * start row and numRows is 25, rows 0 through 24 inclusive
   * will be formatted and returned.
   * @return A String representation of the requested rows.
   */
  public String toString(int startRow, int postindex) {
    return editRows.fromRows(startRow, postindex);
  }

  /** Return one row */
  public String toString(int row) {
    String result = null;
    if (-1 == row) {
      result = editRows.messageRow(" START OF DATA ");
    }
    else if (getNumberOfRows() == row) {
      result = (editRows.messageRow(" END OF DATA "));
    }
    else {
      result = editRows.fromRow(row);
    }
    return result;
  }

  /** Return a string representing the current window view of the data */
  public String toDisplayString() {
    int postindex =  java.lang.Math.min(topRowDisplayed + numRowsInView, getNumberOfRows() + 1);
    return toString (topRowDisplayed, postindex);
  }

  /** Safely set the top row value.
   * Used once the file member is open (and read, if necessary).
   */
  public void setTopRowDisplayed (int val) {
    topRowDisplayed = java.lang.Math.max(-1, java.lang.Math.min(val, getNumberOfRows()));
  }

  /** Fetch index of top row being displayed,
  ranges from -1 to number of records. */
  public int getTopRowDisplayed() {
    return topRowDisplayed;
  }

  /** Return length of SRCDTA field */
  public int getSRCDTAlength() {
    return editRows.getSRCDTAlength();
  }

  /** Return contents of a specific row */
  public String getSRCDTAcontents(int rownum) {
    return editRows.getSRCDTAcontents(rownum);
  }

  /** Safely set the rows-in-view value.
   */
  public void setNumRowsInView (int val) {
    numRowsInView = java.lang.Math.max(0,val);
  }

  /** Fetch number of rows being displayed in the window view. */
  public int getNumRowsInView () {
    return numRowsInView;
  }

  /** Fetch number of Columns actually displayed */
  public int getNumColumnsInView() {
    return editRows.getNumberOfDisplayColumns();
  }

  /** Scroll the view forwards (+) or back (-).
   * Controller should have called setTopRowDisplayed
   * and setNumRowsInView before calling this.
   */
  public String scrollView(int delta) {
    scrollTopOfView(delta);
    return toDisplayString();
  }

  public void scrollTopOfView(int delta) {
    setTopRowDisplayed
    (java.lang.Math.max(-1, java.lang.Math.min(getNumberOfRows(), topRowDisplayed + delta)));
  }

  /** Scroll top of view ahead one page */
  public void scrollTopOfViewAhead() {
    scrollTopOfView(numRowsInView);
  }

  /** Scroll top of view ahead one page */
  public void scrollTopOfViewBack() {
    scrollTopOfView(-numRowsInView);
  }

  /** Scroll top of view ahead one page */
  public void scrollAhead() {
    scrollView(numRowsInView);
  }

  /** Scroll back one page */
  public String scrollBack() {
    return scrollView(-numRowsInView);
  }

  /*****************/
  /* Record Access */
  /*****************/

  /** Load in the file member records to be edited.
   * return true on success. in any event, reset changed flag
   */
  public boolean load() {
    resetDirty();
    return editRows.readAll();
  }

  /** Return the record format or null if it couldn't be gotten. */
  public com.ibm.as400.access.RecordFormat getRecordFormat() {
    return editRows.getRecordFormat();
  }

  /** Cause the record format to be fetched */
  public boolean setRecordFormat() {
    return editRows.setRecordFormat();
  }

  /** Set the record format to match the source edit buffer's format */
  public boolean setRecordFormat (MEUEditBuffer srcBuffer) {
    boolean result = false;
    try {
      result = editRows.setRecordFormat (srcBuffer.getRecordFormat());
    }
    catch (java.beans.PropertyVetoException e) {
      e.printStackTrace();
      result = false;
    }
    return result;
  }

  /** Update a record by row (== record in this case) number. */
  public boolean updateRow (int pos, String contents) {
    boolean result = editRows.updateRow(pos, contents);
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Insert a record row (== record in this case) number. */
  public boolean insertRow (int pos, String contents) {
    boolean result = editRows.insertRow(contents, pos);
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Insert an empty row */
  public boolean insertEmptyRow(int pos) {
    return insertRow(pos, "");
  }

  /** Delete a row. */
  public boolean removeRow (int pos) {
    boolean result = editRows.removeRow(pos);
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Append an empty row. */
  public boolean appendEmptyRow () {
    boolean result = appendRow("");
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Append a row. */
  public boolean appendRow (String contents) {
    boolean result = editRows.appendRow(contents);
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Empty this buffer, change its format, and replace all rows from source */
  public boolean replaceAllRowsWith(MEUEditBuffer sourceBuff) {
    boolean result = editRows.replaceWithAllFrom(sourceBuff.getEditRows());
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Insert all rows from another buffer into this buffer below current row */
  public boolean insertAllRowsFrom(int rownum, MEUEditBuffer sourceBuff) {
    boolean result = editRows.insertAllRowsFrom(rownum, sourceBuff.getEditRows());
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Split/join a row and its follower */
  public boolean splitJoinRow (int rownum, int offset, int lastrow) {
    boolean result = true;
    String rowContents = getSRCDTAcontents(rownum);
    if (lastrow == rownum) {
      if (offset < getSRCDTAlength()) {
        appendRow (rowContents.substring(offset, getSRCDTAlength()));
        updateRow(rownum, rowContents.substring(0, offset));
      }
      else {
        appendEmptyRow();
      }
    }
    else { /* We are not on the last line */
      if (getSRCDTAlength() == offset) { /* At tail of line? Add after. */
        insertEmptyRow(rownum+1);
      }
      else if (0 == offset) { /* At head of line? Push currow down. */
        if (rowContents.trim().equals("")) { /** Head of empty row */
          removeRow(rownum);   /* Slurp next upwards */
        }
        else { /* At head of non-empty row */
          insertEmptyRow(rownum); /* Push down */
        }
      }
      else { /* We are somewhere withing a non-last line */
        if (rowContents.substring(offset).trim().equals("")) {
          /* If blank after offset, join row */
          joinRow(rownum, offset, rowContents);
        }
        else { /* Non-blank after offset, split row */
          splitRow(rownum, offset, rowContents);
        }
      }
    }
    return result;
  }

  /** Split a row at the offset */
  private void splitRow(int rownum, int offset, String rowContents) {
    updateRow(rownum, rowContents.substring(0, offset));
    insertRow(rownum + 1, rowContents.substring(offset));
  }

  /** Join a row at the offset */
  private void joinRow(int rownum, int offset, String rowContents) {
    String nextRowContents = getSRCDTAcontents(rownum + 1);
    rowContents = rowContents.substring(0, offset) + nextRowContents.substring(0, getSRCDTAlength() - offset);
    nextRowContents = nextRowContents.substring(getSRCDTAlength() - offset, getSRCDTAlength());
    updateRow(rownum, rowContents);
    if (nextRowContents.trim().equals("")) {
      removeRow(rownum + 1);
    }
    else {
      updateRow(rownum + 1, nextRowContents);
    }
  }

  /** Copy a row to before another row */
  public boolean copyRowBefore (int fromPos, int followerPos) {
    boolean result = editRows.copyRowBefore(fromPos, followerPos);
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Move a record before a specific position. */
  public boolean moveRowBefore (int fromPos, int followerPos) {
    boolean result = editRows.moveRowBefore(fromPos, followerPos);
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Copy a record after a specific position. */
  public boolean copyRowAfter (int fromPos, int precederPos) {
    boolean result = editRows.copyRowAfter(fromPos, precederPos);
    if (result) {
      setDirty();
    }
    return result;
  }

  /** Move a record after a specific position. */
  public boolean moveRowAfter (int fromPos, int precederPos) {
    boolean result = editRows.moveRowAfter(fromPos, precederPos);
    if (result) {
      setDirty();
    }
    return result;
  }

  /**********************/
  /* Loading and Saving */
  /**********************/

  /** Write out if modified or if forced and set unmodified. */
  public boolean write() {
    return write(false);
  }

  /** Write out if modified and set unmodified. */
  public boolean write(boolean forceWrite) {
    boolean wasWritten = false;
    if (dirty || forceWrite) {
      wasWritten = editRows.writeAll();
      if (wasWritten) {
        resetDirty();
      }
    }
    return wasWritten;
  }

  /** Write out the file member locally to a text stream */
  public void writeLocaltext(java.io.OutputStream o)
  throws java.io.IOException
  {
    o.write(recordsToString().getBytes());
  }

  /** Close one way or another */
  public void closeSession(boolean writeOut) {
    if (writeOut) {
      closeWriting();
    }
    else {
      closeDiscarding();
    }
  }

  /**********************/
  /* File Member Access */
  /**********************/

  /* Create this member with a description */
  public boolean createMember (String description) {
    return editRows.createMember(description);
  }

  /** Return all records as a string */
  public String recordsToString() {
    return editRows.recordsToString();
  }

  /*************************/
  /* Searching & Replacing */
  /*************************/

  /** Returns the rownum on which the next occurrence of
   * a string is found. Caller looks for match on same
   * row where search started (rownum - 1) without invoking
   * this function. This function only used for searching
   * rownum and the following rows.
   */
  public int search (String s, int rownum) {
    int foundOnRow = -1; // Default: not found.
    String thisRowContents = "";
    for (int i = rownum; i <getNumberOfRows(); i++) {
      thisRowContents = getSRCDTAcontents(i);
      if(-1 != thisRowContents.indexOf(s)) {
        foundOnRow = i;
      }
      // System.out.println("This line == " + thisRowContents);
      // System.out.println("This line number == " + i);
      // System.out.println("Found == " + foundOnRow);
      if (-1 != foundOnRow) { // Get out if found.
        break;
      }
    }
    return foundOnRow;
  }

  /** Change next occurence after offset of sought to changed in line rownum.
   * Return new cursor offset or -1 if no change.
   */
  public int changeNextOnLine (String sought, String changed, int rownum, int offset) {
    int result = -1;
    StringBuffer thisRowContents = new StringBuffer(getSRCDTAcontents(rownum));
    StringBuffer contentsFromOffset = new StringBuffer(thisRowContents.substring(offset));
    int foundAt = contentsFromOffset.toString().indexOf(sought);
    int foundEnd = foundAt + sought.length();
    if (-1 != foundAt) {
      // Compose tail of row.
      contentsFromOffset = new StringBuffer(contentsFromOffset.substring(0, foundAt) + changed + contentsFromOffset.substring(foundEnd));
      // Concat head of row with tail.
      String concat = thisRowContents.substring(0, offset) + contentsFromOffset.toString();
      // Truncate if necessary.
      if (concat.length() > thisRowContents.length()) {
        concat = concat.substring(0, thisRowContents.length());
      }
      thisRowContents = thisRowContents.replace(0, concat.length(), concat);
      if (concat.length() < thisRowContents.length()) { // Blank rest of row
        for (int i = thisRowContents.length() - (thisRowContents.length() - concat.length()); i < thisRowContents.length(); i++) {
          thisRowContents.replace(i, i+1, " ");
        }
      }
      if (updateRow(rownum, thisRowContents.toString())) { // Success! Calc new cursor offset
        result = java.lang.Math.min(thisRowContents.length(), offset + foundAt + changed.length());
      }
    }
    return result;
  }

  /***********/
  /* Closing */
  /***********/

  /** Close without bothering to write out. */
  private void closeDiscarding() {
    editRows.close();
  }

  /** Close and write out if necessary */
  private void closeWriting () {
    write();
    closeDiscarding();
  }
}
