/*
 * MEUEditRow.java
 * This is the collection of records an MEU edit session is editing.
 * Created on July 29, 2000, 1:46 PM
 * This is free open source software distributed under the IBM Public License found
 * on the World Wide Web at http://oss.software.ibm.com/developerworks/opensource/license10.html
 * Copyright *C* 2000, Jack J. Woehr, PO Box 51, Golden, CO 80402-0051 USA jax@well.com
 * Copyright *C* 2000, International Business Machines Corporation and others. All Rights Reserved.
 */

package com.SoftWoehr.JTOpenContrib.MEU;

/** Class originally intended to represent one row of the document,
 * now mostly consists of methods for the border between
 * the "record" metaphor and the "row" being edited in unicode in
 * a hightly stream-oriented document architecture.
 * @author jax
 * @version 1.0
 */
public class MEUEditRow extends Object {

  private final MEUFileMember fileMember;

  private static final int lineAreaLength = "0000.00".length();
  private static final int dataAreaLength = "000000".length();

  /** Always associated with one file member */
  MEUEditRow(MEUFileMember fm) {
    fileMember = fm;
  }

  /** Get path of file member */
  public String getIFSPath() {
    return getFileMember().getIFSPath();
  }

  /** Load all rows from the file member on server.
   * Returns true if load appears successful
   */
  public boolean readAll() {
    boolean result = true;
    java.lang.Exception x = null;

    try {
      fileMember.readAll();
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      x = e;
      e.printStackTrace();
      result = false;
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      x = e;
      e.printStackTrace();
      result = false;
    }
    catch (InterruptedException e) {
      x = e;
      e.printStackTrace();
      result = false;
    }
    catch (java.io.IOException e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }
    catch (java.beans.PropertyVetoException e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }

    if (null != x) {
      MEUStockDialog.showExceptionMessageDialog
      (null, "MEU encountered an exception reading all edit rows from " + getFileMember().getIFSPath(), x);
    }
    return result;
  }

  /** Create the new member. We do not create files,
   * ergo, the record format should always be obtainable
   * from the extant file.
   */
  public boolean createMember (String description) {
    boolean result = false;
    try {
      fileMember.createMember (getRecordFormat(), description);
      result = true;
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      e.printStackTrace();
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      e.printStackTrace();
    }
    catch (InterruptedException e) {
      e.printStackTrace();
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
    }
    catch (java.beans.PropertyVetoException e) {
      e.printStackTrace();
    }
    return result;
  }

  /** Clear member on the server and
   * write all rows out fresh into empty member.
   */
  public boolean writeAll() {
    boolean result = true;
    java.lang.Exception x = null;

    try {
      fileMember.writeAllClearingMember();
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }
    catch (InterruptedException e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }
    catch (java.io.IOException e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }
    catch (java.beans.PropertyVetoException e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }
    catch (com.ibm.as400.access.ErrorCompletingRequestException e) {
      x = e;
      // e.printStackTrace();
      result = false;
    }

    if (null != x) {
      MEUStockDialog.showExceptionMessageDialog
      (null, "MEU encountered an exception writing all edit rows to " + getFileMember().getIFSPath(), x);
    }

    return result;
  }

  /** Return the filemember associated with this object. */
  public MEUFileMember getFileMember() {
    return fileMember;
  }

  /** Cause the record format to be fetched */
  public boolean setRecordFormat() {
    return fileMember.setRecordFormat();
  }

  /** Set record format if not open, also clearing records.
   * Return false if open.
   */
  public boolean setRecordFormat (com.ibm.as400.access.RecordFormat rf)
  throws java.beans.PropertyVetoException
  {
    return fileMember.setRecordFormat(rf);
  }

  /** Return the record format or null if it couldn't be gotten. */
  public com.ibm.as400.access.RecordFormat getRecordFormat() {
    com.ibm.as400.access.RecordFormat rf = null;
    try {
      rf = fileMember.getRecordFormat();
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      e.printStackTrace();
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      e.printStackTrace();
    }
    catch (InterruptedException e) {
      e.printStackTrace();
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
    }
    catch (java.beans.PropertyVetoException e) {
      e.printStackTrace();
    }
    return rf;
  }

  /** Return length of SRCDTA field */
  public int getSRCDTAlength() {
    return fileMember.getSRCDTAlength();
  }

  /** Return the contents of a SRCDTA field */
  public String getSRCDTAcontents(int rownum) {
    return fileMember.getSRCDTAcontents(rownum);
  }

  /** Get number of records, really */
  public int getNumberOfRows() {
    return fileMember.getNumberOfRecords();
  }

  /** Return number of columns used to display a row including sequence and date */
  public int getNumberOfDisplayColumns() {
    return lineAreaLength + 1 + getSRCDTAlength() + 1 + dataAreaLength;
  }

  /** Return all records as a string */
  public String recordsToString() {
    return fileMember.recordsToString();
  }

  /** Create a displayable string from the record contents, null if exception caught. */
  private static String fromRecord(com.ibm.as400.access.Record record) {
    /* Create a string buffer the string length plus 3 (one dot and two spaces!) */
    String s = "";
    StringBuffer result = null;
    String retval = null;

    if (null != record) {
      result = new StringBuffer("0000.00 ");
      try {
        s = MEUFileMember.getSRCSEQ(record).toString();
        result.replace(7 - s.length(), 7, s);
        result.append(record.getField(MEUFileMember.SRCDTA_INDEX).toString() + " 000000");
        s = record.getField(MEUFileMember.SRCDAT_INDEX).toString();
      }
      catch (java.io.UnsupportedEncodingException e) {
        e.printStackTrace();
      }
      result.replace(result.length() - s.length(), result.length() + 1, s);
      retval = result.toString();
    }
    return retval;
  }

  /** Get the contents of a row in a form suitable for editing. */
  public String fromRow (int pos) {
    pos = java.lang.Math.max(0, java.lang.Math.min(getNumberOfRows() - 1, pos));
    return fromRecord(fileMember.getRecordAt(pos));
  }

  /** Return a **** MESSAGE **** row. */
  public String messageRow (String msg) {
    StringBuffer result = new StringBuffer ();
    int width = getSRCDTAlength();
    /* debug */
    // System.err.println("SRCDTAlength() == " + Integer.toString(width));
    for (int i = 0; i < width; i++) {
      result.append('*');
    }
    // Center message in result.
    int startpos = java.lang.Math.max(0, ((width / 2) - (msg.length() / 2)) - 1);
    int endpos   = java.lang.Math.min(result.length(), startpos + msg.length());
    result.replace (startpos, endpos,  msg);
    return "        " /* The sequence field and space */
    + result.toString()
    + "       "  /* The date field and space; */
    ;
  }

  /** Make a string equivalent to all the rows desired,
   * appending newlines to each but the last.
   * Starting from -1 means start from "Start of Data" header.
   * Ending with total number of records + 1 means end with "End of Data" footer.
   */
  public String fromRows (int pos, int postindex) {
    StringBuffer result = new StringBuffer();
    pos       = java.lang.Math.max(-1, java.lang.Math.min(getNumberOfRows() - 1, pos));
    postindex = java.lang.Math.max(pos, java.lang.Math.min(getNumberOfRows() + 1, postindex));
    for (; pos < postindex; pos++) {
      if (-1 == pos) {
        result.append(messageRow(" START OF DATA "));
      }
      else if (getNumberOfRows() == pos) {
        result.append(messageRow(" END OF DATA "));
      }
      else {
        result.append(fromRow(pos));
      }
      if (pos != postindex - 1) {
        result.append('\n');
      }
    }
    return result.toString();
  }

  /** Make the big decimal needed to create a SRCDAT timestamp (YYMMDD) */
  public java.math.BigDecimal calcSRCDAT () {
    java.util.GregorianCalendar greg = new java.util.GregorianCalendar();
    String s = Integer.toString(greg.get(java.util.Calendar.YEAR));
    StringBuffer datestamp = new StringBuffer(s.substring(s.length() - 2) + "0000");
    s = Integer.toString(greg.get(java.util.Calendar.MONTH) + 1); /* Month zero-based! */
    datestamp.replace(4 - s.length(), 4, s);
    s = Integer.toString(greg.get(java.util.Calendar.DATE));
    datestamp.replace(6 - s.length(), 6, s);
    /* Debug */
    // System.out.println("New calc'ed date = " + (new java.math.BigDecimal (datestamp.toString())).toString());
    return new java.math.BigDecimal (datestamp.toString());
  }

  /** Calcs next srcseq by 1.00 increments */
  public java.math.BigDecimal calcSRCSEQ(int pos) {
    pos = java.lang.Math.max(0, java.lang.Math.min(getNumberOfRows(), pos));
    java.math.BigDecimal srcseq = new java.math.BigDecimal("0.00");
    if (0 != pos) { // If zeroeth record, just the 0.00 plus increment at end will do.
      try { // If not zeroeth record, look a preceding record's srcseq
        srcseq = fileMember.getSRCSEQ(fileMember.getRecordAt(pos - 1));
      }
      catch (java.io.UnsupportedEncodingException e) {
        e.printStackTrace();
      }
    }
    return srcseq.add(fileMember.ONE_POINT_OH_OH);
  }

  /** Empty and replace with copies of all the rows from another MEUEditRow object.
   * PRESUMES IDENTICAL RECORD FORMAT before call.
   */
  public boolean replaceWithAllFrom(MEUEditRow rows) {
    boolean result = true;
    fileMember.removeAllRecords();
    for (java.util.Enumeration e = rows.getFileMember().elements(); e.hasMoreElements(); ) {
      appendCopyOf((com.ibm.as400.access.Record ) e.nextElement());
    }
    return result;
  }

  public boolean insertAllRowsFrom(int rownum, MEUEditRow rows) {
    boolean result = true;
    for (java.util.Enumeration e = rows.getFileMember().elements(); e.hasMoreElements(); ) {
      // System.out.println ("Row number in insertAllRowsFrom() is " + rownum);
      insertCopyOf((com.ibm.as400.access.Record ) e.nextElement(), rownum++);
    }
    return result;
  }

  /** Append a copy of a record a last record. Update srcseq and srcdat. */
  public boolean appendCopyOf(com.ibm.as400.access.Record sourceRecord) {
    boolean result = false;
    com.ibm.as400.access.Record record = null;
    try {
      record = fileMember.getNewRecord(calcSRCSEQ(getNumberOfRows()), calcSRCDAT(), fileMember.getSRCDTA(sourceRecord));
      result = true;
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      e.printStackTrace();
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      e.printStackTrace();
    }
    catch (InterruptedException e) {
      e.printStackTrace();
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
    }
    catch (java.beans.PropertyVetoException e) {
      e.printStackTrace();
    }

    if (result) {
      fileMember.appendRecord(record);
    }
    return result;
  }

  /** Append a record to the file. Update srcseq and srcdat. */
  public boolean appendRow (com.ibm.as400.access.Record record) {
    java.math.BigDecimal srcseq = calcSRCSEQ(getNumberOfRows());
    fileMember.setSRCSEQ(record, calcSRCDAT());
    fileMember.appendRecord(record);
    return true;
  }

  /** Append a record to the file with srcdta contents. Update srcseq and srcdat. */
  public boolean appendRow (String contents) {
    boolean result = false;
    com.ibm.as400.access.Record record = null;
    java.math.BigDecimal srcseq = calcSRCSEQ(getNumberOfRows());

    try {
      record = fileMember.getNewRecord(srcseq, calcSRCDAT(), contents);
      result = true;
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      e.printStackTrace();
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      e.printStackTrace();
    }
    catch (InterruptedException e) {
      e.printStackTrace();
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
    }
    catch (java.beans.PropertyVetoException e) {
      e.printStackTrace();
    }

    if (result) {
      fileMember.appendRecord(record);
    }
    return result;
  }

  /** Insert a copy of a record at the position indicated, pushing all after back. */
  public boolean insertCopyOf(com.ibm.as400.access.Record sourceRecord, int pos) {
    boolean result = false;
    try {
      pos = java.lang.Math.max(0, java.lang.Math.min(getNumberOfRows(), pos));
      java.math.BigDecimal srcseq = calcSRCSEQ(pos);
      fileMember.resequenceFrom(pos, getNumberOfRows(), srcseq.add(fileMember.ONE_POINT_OH_OH), fileMember.ONE_POINT_OH_OH);
      com.ibm.as400.access.Record record = fileMember.getNewRecord(srcseq, calcSRCDAT(), fileMember.getSRCDTA(sourceRecord));
      fileMember.insertRecord(record, pos);
      result = true;
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      e.printStackTrace();
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      e.printStackTrace();
    }
    catch (InterruptedException e) {
      e.printStackTrace();
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
    }
    catch (java.beans.PropertyVetoException e) {
      e.printStackTrace();
    }
    catch (ArrayIndexOutOfBoundsException e) {
      e.printStackTrace();
    }
    return result;
  }

  /** Create a new row at pos using record as the source */
  public boolean insertRow (com.ibm.as400.access.Record record, int pos) {
    boolean result = false;
    pos = java.lang.Math.max(0, java.lang.Math.min(getNumberOfRows(), pos));
    java.math.BigDecimal srcseq = calcSRCSEQ(pos);
    fileMember.setSRCSEQ(record, calcSRCDAT());
    fileMember.resequenceFrom(pos, getNumberOfRows(), srcseq.add(fileMember.ONE_POINT_OH_OH), fileMember.ONE_POINT_OH_OH);
    fileMember.insertRecord(record, pos);
    result = true;
    return result;
  }

  /** Insert a new row into the records of the MEUFileMember
   * based on a string to be the data contents.
   */
  public boolean insertRow( String row, int pos) {
    boolean result = true;
    try {
      pos = java.lang.Math.max(0, java.lang.Math.min(getNumberOfRows(), pos));
      java.math.BigDecimal srcseq = calcSRCSEQ(pos);
      fileMember.resequenceFrom(pos, getNumberOfRows(), srcseq.add(fileMember.ONE_POINT_OH_OH), fileMember.ONE_POINT_OH_OH);
      com.ibm.as400.access.Record record = fileMember.getNewRecord(srcseq, calcSRCDAT(), row);
      fileMember.insertRecord(record, pos);
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      e.printStackTrace();
      result = false;
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      e.printStackTrace();
      result = false;
    }
    catch (InterruptedException e) {
      e.printStackTrace();
      result = false;
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
      result = false;
    }
    catch (java.beans.PropertyVetoException e) {
      e.printStackTrace();
      result = false;
    }
    catch (ArrayIndexOutOfBoundsException e) {
      e.printStackTrace();
      result = false;
    }
    return result;
  }

  /** Give new contents to a row.
   * Handle last modified date.
   * @return true if successful.
   */
  public boolean updateRow (int pos, String contents) {
    boolean result = false;
    com.ibm.as400.access.Record record = fileMember.getRecordAt(pos);
    if (null != record) {
      fileMember.setSRCDTA (record, contents);
      fileMember.setSRCDAT (record, calcSRCDAT());
      result = true;
    }
    return result;
  }

  /** Delete a row. */
  public boolean removeRow (int pos) {
    boolean result = false;
    // This gets incremented in the resequence call.
    java.math.BigDecimal srcseq = new java.math.BigDecimal("0.00");
    try {

      // Maybe this test shouldn't be here. May be it prevents an exception
      // from being thrown that should be. But we *do* have to test for pos>0
      // so we know if to get a SRCSEQ other than 0001.00
      if (pos > 0 && pos < getNumberOfRows()) {
        srcseq = fileMember.getSRCSEQ(fileMember.getRecordAt(pos - 1));
      }
      fileMember.removeRecord(pos);
      result = true;
    }
    catch (ArrayIndexOutOfBoundsException e) {
      e.printStackTrace();
    }
    catch (java.io.UnsupportedEncodingException e) {
      e.printStackTrace();
    }
    if (result) // Resequence .iff. success
    fileMember.resequenceFrom(pos, getNumberOfRows(), srcseq.add(fileMember.ONE_POINT_OH_OH), fileMember.ONE_POINT_OH_OH);
    return result;
  }

  /** Copy a row to before another row */
  public boolean copyRowBefore (int fromPos, int followerPos) {
    boolean result = false;
    com.ibm.as400.access.Record record = fileMember.getRecordAt(fromPos);
    result = insertCopyOf(record, followerPos);
    return result;
  }

  /** Move a record before a specific position. */
  public boolean moveRowBefore (int fromPos, int followerPos) {
    boolean result = false;
    result = copyRowBefore(fromPos, followerPos);
    if (fromPos >= followerPos)
    { // fromPos++ because record to delete got pushed ahead
      fromPos++;
    }
    if (result) {
      result = removeRow(fromPos);
    }
    return result;
  }

  /** Copy a record after a specific position. */
  public boolean copyRowAfter (int fromPos, int precederPos) {
    return copyRowBefore(fromPos, ++precederPos);
  }

  /** Move a record after a specific position. */
  public boolean moveRowAfter (int fromPos, int precederPos) {
    return moveRowBefore(fromPos, ++precederPos);
  }

  /*
  public void copyRangeBefore (int firstPos, int lastPos, int followerPos) {
  }

  public void moveRangeBefore (int fromPos,  int lastPos, int followerPos) {
  }

  public void copyRangeAfter (int fromPos,  int lastPos, int precederPos) {
  }

  public void moveRangeAfter (int fromPos,  int lastPos, int precederPos) {
  }
  */

  /** Close and discard */
  public void close() {
    try {
      fileMember.close();
    }
    catch (com.ibm.as400.access.AS400Exception e) {
      e.printStackTrace();
    }
    catch (com.ibm.as400.access.AS400SecurityException e) {
      e.printStackTrace();
    }
    catch (InterruptedException e) {
      e.printStackTrace();
    }
    catch (java.io.IOException e) {
      e.printStackTrace();
    }
  }

  /**********************/
  /* Some test routines */
  /**********************/

  /** Read in rows, display, modify, and write out */
  public static void testWrite (String server, String lib, String file, String member) {

    // Create object representing extant file member.
    MEUFileMember fm = new MEUFileMember(new com.ibm.as400.access.AS400(server), lib, file, member);

    // Create an edit row
    MEUEditRow editRows = new MEUEditRow(fm);

    // Get its record format.
    com.ibm.as400.access.RecordFormat rf = editRows.getRecordFormat();

    if (null == rf) {
      System.out.println("Couldn't get record format.");
    }

    else {
      editRows.getFileMember().displayRecordFormat();

      // Read all in.
      editRows.readAll();

      System.out.println("Print out all rows:");
      System.out.println(editRows.fromRows(-1, editRows.getNumberOfRows() + 1));
      System.out.println("Print out all records:");
      System.out.print(fm.recordsToString());

      System.out.println("Inserting a row.");
      /** Insert a row */
      editRows.insertRow("This tests our ability to insert a row at 4.", 4);
      System.out.println("Redisplaying with inserted record.");

      System.out.println("Print out all rows (with modification) :");
      System.out.println(editRows.fromRows(-1, editRows.getNumberOfRows() + 1));

      System.out.println("===");
      System.out.println("Now trying to write back all records due to one appended record.");
      /** Write out everything, clearing SRCPFM first. */
      editRows.writeAll();
      System.out.println("===");

      /* The above writeAll() instantiates afresh the record array, so now we can print out the records. */
      System.out.println("Print out all records:");
      System.out.print(fm.recordsToString());

      System.out.println("===");
      System.out.println("All rows were written.");
      System.out.println("===");

    }
  }

  /** Read in rows and display */
  public static void testRead (String server, String lib, String file, String member) {

    // Create object representing extant file member.
    MEUFileMember fm = new MEUFileMember(new com.ibm.as400.access.AS400(server), lib, file, member);

    // Create an edit row
    MEUEditRow editRows = new MEUEditRow(fm);

    // Get its record format.
    com.ibm.as400.access.RecordFormat rf = editRows.getRecordFormat();

    if (null == rf) {
      System.out.println("Couldn't get record format.");
    }

    else {
      editRows.getFileMember().displayRecordFormat();

      // Read all in.
      editRows.readAll();
      System.out.println("Number of rows: " + Integer.toString(editRows.getNumberOfRows()));
      System.out.println("Print out all rows:");
      System.out.println(editRows.fromRows(-1, editRows.getNumberOfRows() + 1));
      System.out.println("Print out all records:");
      System.out.print(fm.recordsToString());

    }
  }

  /** Just a quick test of record access */
  public static void main (String argv[]) {

    // Make sure we have enuff args.
    if (argv.length < 4) {
      System.out.println ("Usage: MEUEditRow server lib file member");
      System.exit(0);
    }

    // Interpret our args.
    String server = argv[0];
    String lib = argv[1];
    String file = argv[2];
    String member = argv[3];

    String testType = "read";

    if (argv.length >= 5) {
      testType = argv[4].toLowerCase();
    }

    if (testType.equals("read")) {
      testRead(server, lib, file, member);
    }
    else if (testType.equals("write")) {
      testWrite(server, lib, file, member);
    }
    System.out.println(testType + " test completed");
    System.exit(0);
  }
}
