/*
 * MEUTextArea.java
 *
 * This is the text "terminal" for MEU's edit sessions.
 * Created on July 28, 2000, 11:59 PM
 * This is free open source software distributed under the IBM Public License found
 * on the World Wide Web at http://oss.software.ibm.com/developerworks/opensource/license10.html
 * Copyright *C* 2000, Jack J. Woehr, PO Box 51, Golden, CO 80402-0051 USA jax@well.com
 * Copyright *C* 2000, International Business Machines Corporation and others. All Rights Reserved.
 */

package com.SoftWoehr.JTOpenContrib.MEU;

/** The edit region.
 * @author jax
 * @version 1.0
 */
public class MEUTextArea extends javax.swing.JTextArea {

  /** The controller associated with the frame in which this text area appears.
   */
  private MEUController myController;

  /************************************/
  /* Visible Display Region Constants */
  /************************************/

  /** The template string for SRCSEQ */
  private static final String s_srcseq;

  /** The template string for SRCDAT */
  private static final String s_srcdat;

  /** All the width of the constant stuff in a row
   */
  private static final int constant_row_width;

  /** this could be fetched from a property */
  private static final int default_data_width;

  /** Length of this start-of-row field */
  private static final int srcseq_length;

  /** Length of this end-of-row field */
  private static final int srcdat_length;

  /* Offset in a row to start of source data */
  public static final int offset_to_srcdta;

  /** The default number of visible columns. SRCSEQ + ' ' + SRCDTA + ' ' + SRCDAT */
  private static final int DEFAULT_COLUMNS;

  /** Default number of rows, equivalent to a 5250 display. */
  private static final int DEFAULT_ROWS;

  /** Blanks the length of srcseq */
  private static final String blank_line_command;

  /** End of record, e.g., newline */
  private static final String s_eor;

  static {
    s_eor    = "\n";
    s_srcseq = "0000.00";
    s_srcdat = "000000";
    srcdat_length = s_srcdat.length();
    srcseq_length = s_srcseq.length();
    offset_to_srcdta = srcseq_length + 1;
    StringBuffer sb = new StringBuffer(srcseq_length);
    for (int i = 0; i < srcseq_length; i++) {
      sb.append(' ');
    }
    blank_line_command = sb.toString();
    default_data_width= 80;
    DEFAULT_ROWS = 25;
    /* SRCSEQ + ' ' + ' ' + SRCDAT */
    constant_row_width = srcseq_length + 1 + 1 + srcdat_length;

    /* The default number of visible columns. SRCSEQ + ' ' + SRCDTA + ' ' + SRCDAT */
    DEFAULT_COLUMNS = constant_row_width + default_data_width;
  }

  /****************************************************/
  /* Variables describing taxonomony of the text area */
  /****************************************************/

  /** Columns in pane */
  private int myColumns;

  /** Rows in pane */
  private int myRows;

  /** Foreground color */
  private java.awt.Color myForegroundColor = java.awt.Color.green;

  /** Background color */
  private java.awt.Color myBackgroundColor = java.awt.Color.black;

  /** Caret color */
  private java.awt.Color myCaretColor = java.awt.Color.green;

  /** A custom keymap */
  private javax.swing.text.Keymap keymap;

  /** Are we inserting or typeover? */
  private boolean inserting = false;

  /** Line attributes */
  private LineAttributes [] lineAttribs;

  /** Set afresh the line attributes */
  private void initLineAttributes () {
    lineAttribs = new LineAttributes [myRows];
    for (int i = 0; i < myRows; i++) {
      lineAttribs[i] = new LineAttributes();
    }
  }

  /****************/
  /* Constructors */
  /****************/

  /** Create the text area dependend on a frame controller.
   * This is the only runtime-useful ctor.
   */
  public MEUTextArea (MEUController mc) {
    setController(mc);

    /* Fallback defaults */
    int columns = DEFAULT_COLUMNS;
    int rows    = DEFAULT_ROWS;

    /* Get defaults present at least in the default properties object,
     * but might be screwy due to manual changes by user to props file.
     */
    try {
      columns = getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.edittext.colums");
    }
    catch (MEUProperties.NoSuchPropertyException e) {
    }
    catch (MEUProperties.InvalidIntPropertyException e) {
    }

    try {
      columns = getIntProperty("com.SoftWoehr.JTOpenContrib.MEU.window.edittext.colums");
    }
    catch (MEUProperties.NoSuchPropertyException e) {
    }
    catch (MEUProperties.InvalidIntPropertyException e) {
    }

    reinit(columns, rows);
    addCustomKeymap();

    /* Add a focus listener to put our cursor somewhere visible the
     * first time it gets focus.
     */
    addFocusListener (new java.awt.event.FocusListener() {
      public void focusGained(java.awt.event.FocusEvent e) {
        wrapCursor();
      }
      public void focusLost(java.awt.event.FocusEvent e) {
      }
    });


    initLineAttributes(); // Make sure we have line attributes to change.
    setEnabled(false); // Start off disabled until controller enables

    /*
     * addPropertyChangeListener("caretColor", new java.beans.PropertyChangeListener () {
     *   public void propertyChange(java.beans.PropertyChangeEvent evt) {
     *     System.out.println("Caret prop change == " + evt.toString());
     * }
     * });
     */
  }

  /** Creates new MEUTextArea of default size and color.
   * Only used by the design tool (to load a representation of the bean).
   */
  public MEUTextArea() {
    this(DEFAULT_COLUMNS, DEFAULT_ROWS, java.awt.Color.green, java.awt.Color.black);
  }

  /** Creates new MEUTextArea of specific size and specific color */
  public MEUTextArea(int columns,
  int rows,
  java.awt.Color foregroundColor,
  java.awt.Color backgroundColor) {
    reinit(columns,rows,foregroundColor, backgroundColor);
    addCustomKeymap();
    setEnabled(false); // Start off disabled until controller enables
  }

  /** Set the MEUController for this component,
   *  called in ctor, passed in from MEUFrame.
   */
  protected void setController (MEUController controller) {
    myController = controller;
  }

  /***********************/
  /* Visual taxonomy     */
  /***********************/

  /** Useful reinitialization of visual settings */
  public void reinit(int columns, int rows) {
    initFont();
    resetSize(columns, rows);
  }

  /** Useful reinitialization of visual settings */
  public void reinit(int columns,
  int rows,
  java.awt.Color foregroundColor,
  java.awt.Color backgroundColor) {
    reinit(columns, rows);
    resetColor(foregroundColor, backgroundColor);
  }

  /** Useful reinitialization of visual settings */
  public void reinit(int columns,
  int rows,
  java.awt.Color foregroundColor,
  java.awt.Color backgroundColor,
  java.awt.Color caretColor) {
    reinit(columns, rows);
    resetColor(foregroundColor, backgroundColor, caretColor);
  }

  /** Set the rows and columns.
   */
  public void initSize() {
    setColumns(myColumns);
    setRows(myRows);
    setMinimumSize
    (new java.awt.Dimension(myColumns * getColumnWidth(),
    myRows * getRowHeight()));
    setMaximumSize
    (new java.awt.Dimension(myColumns * getColumnWidth(),
    myRows * getRowHeight()));
  }

  /** Resize the text area to a new cols & rows.
   * Probably shouldn't ever do this with or without a changed
   * buffer without doing a check changed and redisplay.
   */
  public void resetSize (int columns, int rows) {
    myColumns = columns;
    myRows    = rows;
    initSize();
    initLineAttributes();
  }

  /** Init the font.
   */
  public void initFont() {
    setFont(new java.awt.Font("Monospaced",java.awt.Font.PLAIN,16));
  }

  /** Change the font and resize */
  public void resetFont (java.awt.Font f) {
    setFont(f);
    initSize();
  }

  /** Reset  the caret color only */
  public void resetColor (java.awt.Color caretColor) {
    myCaretColor = caretColor;
    setCaretColor(myCaretColor);
  }

  /** Set the text colors */
  public void resetColor (java.awt.Color foregroundColor, java.awt.Color backgroundColor) {
    myForegroundColor = foregroundColor;
    myBackgroundColor = backgroundColor;
    setForeground(myForegroundColor);
    setBackground(myBackgroundColor);
  }

  /** Set the text colors and caret color */
  public void resetColor (java.awt.Color foregroundColor, java.awt.Color backgroundColor, java.awt.Color caretColor) {
    resetColor(foregroundColor, backgroundColor);
    resetColor(caretColor);
  }

  /** Necessitated because I can't convince the IDE to stop setting
   * the Caret color in MEUFrame.initComponents() ... arghh!
   */
  public java.awt.Color getCaretColorFromProperty () {
    java.awt.Color crcolor = java.awt.Color.green;
    String cc  = getProperty("com.SoftWoehr.JTOpenContrib.MEU.window.color.cursor");
    if (null != cc) {
      crcolor = toColor(cc);
    }
    return crcolor;
  }
  /** Necessitated because I can't convince the IDE to stop setting
   * the Caret color in MEUFrame.initComponents() ... arghh!
   */
  public java.awt.Color getForegroundColorFromProperty () {
    java.awt.Color fgcolor = java.awt.Color.green;
    String cc  = getProperty("com.SoftWoehr.JTOpenContrib.MEU.window.color.foreground");
    if (null != cc) {
      fgcolor = toColor(cc);
    }
    return fgcolor;
  }
  /** Necessitated because I can't convince the IDE to stop setting
   * the Caret color in MEUFrame.initComponents() ... arghh!
   */
  public java.awt.Color getBackgroundColorFromProperty () {
    java.awt.Color bgcolor = java.awt.Color.black;
    String cc  = getProperty("com.SoftWoehr.JTOpenContrib.MEU.window.color.background");
    if (null != cc) {
      bgcolor = toColor(cc);
    }
    return bgcolor;
  }


  /*************************************/
  /* Geogrpahy and motion in text area */
  /*************************************/

  /** Return total number of chars actually in the text area contents */
  private int getBufferSize() {
    // Plus-one for the '\n' on each row EXCEPT the last.
    return getText().length();
  }

  /** Get number of rows actually in buffer */
  private int getNumberOfRows() {
    // return getBufferSize() / rowLengthWithNewline(); /* This is wrong because last row has no newline! */
    return myRows;
  }

  /* */
  private int rowLengthVisible() {
    return myColumns;
  }

  /* */
  private int rowLengthWithNewline () {
    return rowLengthVisible() + 1;
  }

  /** Returns the offset of start-of-row (at SRCSEQ field)
   * within the text contents.
   */
  private int getRowOffset(int rownum) {
    return rownum * rowLengthWithNewline();
  }

  /** Get offset from head of row of the current position */
  public int getOffsetInRow(int curpos) {
    return curpos % rowLengthWithNewline();
  }

  /** Return an offset within SRCDTA, flooring and ceiling-ing it. */
  public int getOffsetInSRCDTA(int curpos) {
    int offset = getOffsetInRow(curpos) - offset_to_srcdta;
    return java.lang.Math.min(java.lang.Math.max(0, offset), getSRCDTAlength());
  }

  /** Move cursor in current line but always ends up within srcdta floor/ceiling. */
  public void setOffsetInSRCDTA(int offset) {
  int pos = getRowOffsetFromCurpos(getCurPos()) + offset_to_srcdta;
  pos += java.lang.Math.max(0, java.lang.Math.min(getSRCDTAlength(), offset));
  setCurPos(pos);
  }

  /** Get offset of head of row from current position */
  private int getRowOffsetFromCurpos(int curpos) {
    return (curpos / rowLengthWithNewline()) * rowLengthWithNewline();
  }

  /** Convert cursor position to row number it resides on. */
  private int getRowNumber (int curpos) {
    return curpos / rowLengthWithNewline();
  }

  /** Return the number of the row in the display the cursor is on */
  public int getDisplayedRowNumber () {
    return getRowNumber(getCaretPosition());
  }

  /** Increment curpos at head of a row to be at start of srcdta */
  private int toStartOfSrcdta (int currentCursorPos) {
    return currentCursorPos + offset_to_srcdta;
  }

  /** Increment curpos at head of a row to be at end of srcdta */
  private int toEndOfSrcdta(int currentCursorPos) {
    /* An alternative way of inferring this location without calling getSRCDTAlength() */
    return currentCursorPos + rowLengthVisible() - (1 + srcdat_length);
  }

  /** Increment curpos at head of a row to be at end of srcseq */
  private int toEndOfSrcseq(int currentCursorPos) {
    /* An alternative way of inferring this location without calling getSRCDTAlength() */
    return currentCursorPos + srcseq_length;
  }

  /** Return length of SRCDTA field */
  public int getSRCDTAlength() {
    return myController.getSRCDTAlength();
  }

  /** Get contents of SRCDTA field */
  public String getSRCDTAcontents(int rownum) {
    String result = null;
    try {
      result = getText(toStartOfSrcdta(getRowOffset(rownum)), getSRCDTAlength());
    }
    catch (javax.swing.text.BadLocationException e) {
      e.printStackTrace();
    }
    return result;
  }

  /** Removing trailing blanks */
  public String minusTrailing (String s) {
    int len = s.length();
    for (; len > 0 ; len--) {
      if (s.charAt(len-1) != ' ')
      {
        break;
      }
    }
    return s.substring(0, len);
  }

  /** Length of text before all-blanks at tail of line */
  private int lengthOfSrcdtaText(int currentCursorPos) {
    int currowaddr = getRowOffsetFromCurpos(currentCursorPos);
    int result = toEndOfSrcdta(currowaddr); // If exception, harmlessly to end of SRDTA field
    try {
      result = minusTrailing(getText(toStartOfSrcdta(currowaddr), getSRCDTAlength())).length();
    }
    catch (javax.swing.text.BadLocationException e) {
      e.printStackTrace();
    }
    return result;
  }

  /** Increment curpos at head of a row to be at end of srcdta text before blank trailers */
  private int toEndOfSrcdtaText(int currentCursorPos) {
    return toStartOfSrcdta(currentCursorPos) + lengthOfSrcdtaText(currentCursorPos);
  }

  /** Increment curpos at head of a row to be at end of visible line. */
  private int toEndOfVisibleLine (int currentCursorPos) {
    return currentCursorPos + rowLengthVisible();
  }

  /** See if we need to wrap cursor val by line */
  private boolean atEndOfVisibleLine (int currentCursorPos) {
    if (currentCursorPos % rowLengthWithNewline() == rowLengthVisible()) {
      return true;
    }
    return false;
  }

  /** Cursor val by buffer size */
  private boolean atEndOfBuffer (int currentCursorPos) {
    return currentCursorPos == getBufferSize();
  }

  /** WRap cursor val by buffer size */
  private boolean pastEndOfBuffer (int currentCursorPos) {
    return currentCursorPos >  getBufferSize();
  }

  /** Convert a cursor offset to where the cursor should be,
   * i.e., if it's on the end of the line put it to the next
   * line, if it's at the end of the buffer, wrap to top.
   */
  public int wrappedCursorPos (int currentCursorPos) {
    // First try pushing to next line
    if (atEndOfVisibleLine(currentCursorPos)) {
      currentCursorPos++;
    }
    if (atEndOfBuffer(currentCursorPos)) {
      currentCursorPos = 0;
    }
    // Then see if we are wrapping to top
    if (pastEndOfBuffer(currentCursorPos)) {
      currentCursorPos %= rowLengthWithNewline();
    }

    return java.lang.Math.max(currentCursorPos,0);
  }

  /** Recalc cursor pos */
  private int upCursorPos (int currentCursorPos) {
    int curpos = currentCursorPos - rowLengthWithNewline();
    if (curpos < 0) {
      curpos += getBufferSize() + 1;
    }
    return curpos;
  }

  /** Recalc cursor pos */
  private int downCursorPos (int currentCursorPos) {
    return wrappedCursorPos(currentCursorPos + rowLengthWithNewline());
  }

  /** Recalc cursor pos */
  private int leftCursorPos (int currentCursorPos) {
    int curpos = --currentCursorPos;
    if (atEndOfVisibleLine(curpos)) {
      curpos--;
    }
    if (curpos < 0) {
      curpos = getBufferSize() - 1;
    }
    return curpos;
  }

  /** Recalc cursor pos */
  private int rightCursorPos (int currentCursorPos, int increment) {
    int result = currentCursorPos;
    if (increment >0 ) {
      result = wrappedCursorPos(currentCursorPos+increment);
    }
    return result;
  }

  /** Recalc cursor pos */
  private int rightCursorPos (int currentCursorPos) {
    return wrappedCursorPos(++currentCursorPos);
  }

  /** Recalc cursor pos */
  private int homeCursorPos  (int currentCursorPos) {
    return  toStartOfSrcdta(getRowOffsetFromCurpos(currentCursorPos));
  }

  /** Recalc cursor pos */
  private int endCursorPos (int currentCursorPos) {
    return  toEndOfSrcdtaText(getRowOffsetFromCurpos(currentCursorPos));
  }

  /** Recalc cursor pos */
  private int lineAreaCursorPos (int currentCursorPos) {
    return  getRowOffsetFromCurpos(currentCursorPos);
  }

  /** Recalc cursor pos */
  private int srcdtaEndCursorPos (int currentCursorPos) {
    return  toEndOfSrcdta(getRowOffsetFromCurpos(currentCursorPos));
  }

  /** Recalc cursor pos */
  private int topHomeCursorPos  (int currentCursorPos) {
    return  offset_to_srcdta;
  }

  /** Recalc cursor pos */
  private int bottomEndCursorPos (int currentCursorPos) {
    return  toEndOfSrcdtaText(getRowOffsetFromCurpos(getBufferSize()));
  }

  /** Recalc cursor pos */
  private int pageUpCursorPos (int currentCursorPos) {
    return currentCursorPos - getRowOffsetFromCurpos(currentCursorPos);
  }

  /** Recalc cursor pos */
  private int pageDownCursorPos (int currentCursorPos) {
    return currentCursorPos - getRowOffsetFromCurpos(currentCursorPos) + getRowOffsetFromCurpos(getBufferSize());
  }

  /****************************/
  /* Actually move the cursor */
  /****************************/

  /** Get the cursor position */
  public int getCurPos () {
    return getCaretPosition();
  }

  /** Set the cursor position */
  public boolean setCurPos (int pos) {
    boolean result = true;
    setCaretPosition(pos);
    return result;
  }

  /** Allows frame to make sure cursor is in reasonable position
   * e.g., after buffer is loaded and cursor is at end.
   */
  public void wrapCursor() {
    setCaretPosition(wrappedCursorPos(getCaretPosition()));
  }

  /** Set cursor pos */
  public void moveCursorUp () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(upCursorPos(getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorDown () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(downCursorPos(getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorLeft () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(leftCursorPos(getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());

  /** Set cursor pos */}
  public void moveCursorRight () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(rightCursorPos(getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorHome () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(homeCursorPos(getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorEnd () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(endCursorPos(getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorTopHome() {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(topHomeCursorPos (getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorBottomEnd () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(bottomEndCursorPos (getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorPageUp () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(pageUpCursorPos (getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorPageDown () {
    /* debug */ // System.out.println("Curpos was " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
    setCaretPosition(pageDownCursorPos (getCaretPosition()));
    /* debug */ // System.out.println("Curpos is " + getCaretPosition() + " Cols == " + myColumns + " Rows == " + myRows + " Bufsize " + getBufferSize());
  }

  /** Set cursor pos */
  public void moveCursorLineArea () {
    setCaretPosition(lineAreaCursorPos (getCaretPosition()));
  }

  /** Set cursor pos */
  public void moveCursorSrcdtaEnd () {
    setCaretPosition(srcdtaEndCursorPos (getCaretPosition()));
  }

  /*********************/
  /* Protected Regions */
  /*********************/

  private boolean pastEndOfSrcdta (int pos) {
    int currowaddr = getRowOffsetFromCurpos(pos);
    return pos >= toEndOfSrcdta (currowaddr);
  }

  private boolean inDeadColumn(int pos) {
    int currowaddr = getRowOffsetFromCurpos(pos);
    return pos == currowaddr + srcseq_length;
  }

  private boolean onProtectedRow(int pos) {
    return getRowWriteProtected(getRowNumber(pos));
  }

  private boolean inProtectedRegion (int pos) {
    return pastEndOfSrcdta(pos) | inDeadColumn(pos) | onProtectedRow(pos);
  }

  /** In Line command region ? */
  private boolean inLineCommandRegion (int pos) {
    int currowaddr = getRowOffsetFromCurpos(pos);
    return pos < currowaddr + srcseq_length;
  }

  /********************/
  /* Line attributes */
  /*******************/

  private class LineAttributes {
    private boolean write_protected = false;
    private boolean changed = false;
    private StringBuffer lineCommand;

    /** Init the line command buffer to blanks the length of the line command field */
    public LineAttributes() {
      initLineCommand();
    }

    /** Set line command to blanks, correct length */
    public void initLineCommand() {
      lineCommand = new StringBuffer(blank_line_command);
    }

    public boolean isWriteProtected() {
      return write_protected;
    }

    public void setWriteProtected(boolean yesno) {
      write_protected = yesno;
    }

    public boolean isChanged() {
      return changed;
    }

    public void setChanged(boolean changedNotChanged) {
      changed = changedNotChanged;
    }

    /** When the string is blank there is no line command */
    public boolean hasLineCommand() {
      return (lineCommand.toString().trim() != "");
    }

    /** Returns whole string untrimmed. */
    public String getLineCommand() {
      return lineCommand.toString();
    }

    /** Set line command to a string. String must be correct length! (srcseq_length) */
    public void setLineCommand (String c) {
      if (null == c) {
        initLineCommand();
      }
      else {
        lineCommand.replace(0, java.lang.Math.min(c.length(), lineCommand.length()), c);
      }
    }

    /** Set one char of the line command */
    public void setLineCommandChar (int pos, String c) {
      lineCommand.replace(pos, pos+1, c);
    }

    /** Remove a char and backfill with blank. */
    public void deleteLineCommandChar (int pos) {
      lineCommand.replace(pos, lineCommand.length(), lineCommand.substring(pos+1) + " ");
    }

    /** Backspace over a char and backfill with blank. */
    public void backspaceLineCommandChar(int pos) {
      lineCommand.replace(pos-1, lineCommand.length(), lineCommand.substring(pos) + " ");
    }
  }

  /** Set a row write-protected or not */
  public void setRowWriteProtected(boolean yesno, int rownum) {
    lineAttribs[rownum].setWriteProtected(yesno);
  }

  /** Set a row changed or not */
  public void setRowChanged(boolean yesno, int rownum) {
    lineAttribs[rownum].setChanged(yesno);
  }

  /** Set a row line command text */
  public void setRowLineCommand(String c, int rownum) {
    lineAttribs[rownum].setLineCommand(c);
  }

  /** Set one char (first char of string) in one pos in row line command text */
  public void setRowLineCommandChar(int pos, String c, int rownum) {
    lineAttribs[rownum].setLineCommandChar(pos, c);
  }

  public void deleteRowLineCommandChar(int pos, int rownum) {
    lineAttribs[rownum].deleteLineCommandChar(pos);
  }

  public void backspaceRowLineCommandChar(int pos, int rownum) {
    lineAttribs[rownum].backspaceLineCommandChar(pos);
  }
  /** Set a row write-protected or not */
  public boolean getRowWriteProtected(int rownum) {
    return lineAttribs[rownum].isWriteProtected();
  }

  /** get a row changed or not */
  public boolean getRowChanged(int rownum) {
    return lineAttribs[rownum].isChanged();
  }

  /** get a row's line command or null */
  public String getRowLineCommand(int rownum) {
    return lineAttribs[rownum].getLineCommand();
  }

  /** get the row line commands in a vector */
  public java.util.Vector getRowLineCommands() {
    java.util.Vector v = new java.util.Vector(lineAttribs.length);
    for (int i = 0; i < getNumberOfRows(); i++) {
      // /* Debug */ System.out.println ("Getting line command for row " + i);
      v.add(getRowLineCommand(i));
    }
    return v;
  }

  /** Get a row's having a line command */
  public boolean rowHasLineCommand(int rownum) {
    return lineAttribs[rownum].hasLineCommand();
  }

  /**************************/
  /* Terminal functionality */
  /**************************/

  public void setToBlank (int numberOfRows) {
    StringBuffer sb = new StringBuffer(rowLengthVisible());
    for (int i = 0; i < rowLengthVisible(); i++) {
      sb.append(' ');
    }
    setText("");
    for (int i = 0; i < numberOfRows; i++) {
      append(sb.toString());
      if (i < numberOfRows - 1) {
        append("\n");
      }
    }
  }

  /** Replaces visible portion of row. Caller better be right about length! */
  public void replaceVisibleRow
  (int rownum, String text, boolean isProtected, boolean isChanged, String lineCommand) {
    int rowOffset = getRowOffset(rownum);
    replaceRange(text, rowOffset, toEndOfVisibleLine(rowOffset));
    setRowWriteProtected(isProtected, rownum);
    setRowChanged(isChanged, rownum);
    setRowLineCommand(lineCommand, rownum);
  }

  /****************/
  /* Text Editing */
  /****************/

  /** Return insert mode (true) or typeover (false) */
  public boolean getInserting() {
    return inserting;
  }

  /** Toggle the insert mode flag */
  public boolean toggleInserting () {
    return inserting = !inserting;
  }

  /** Replace char at current pos */
  private void typeoverChar (String c) {
    int pos = getCaretPosition();
    replaceRange(c, pos, pos+1);
    setRowChanged(true, getRowNumber(pos));
    setCaretPosition(wrappedCursorPos(++pos));
  }

  /** Insert char at current pos */
  private void insertChar (String c) {
    int pos = getCaretPosition();
    try {
      int currowaddr = getRowOffsetFromCurpos(pos);
      String right_text = "";

      if (toEndOfSrcdtaText(currowaddr) > pos) {
        right_text = minusTrailing(getText(pos, toEndOfSrcdtaText(currowaddr) - pos));
      }

      /* Is any non-blank to right? */
      if (right_text.length() == 0) {
        replaceRange(c, pos, pos + 1);
        setRowChanged(true, getRowNumber(pos));
        setCaretPosition(wrappedCursorPos(++pos));
      }

      /* Musn't shove text off line */
      else if (pos + right_text.length() == toEndOfSrcdta(currowaddr)) {
        beep();
      }
      else { /* It's okay, there's room */
        replaceRange(c + right_text, pos, pos + (c + right_text).length());
        setRowChanged(true, getRowNumber(pos));
        setCaretPosition(wrappedCursorPos(++pos));
      }
    }
    catch (javax.swing.text.BadLocationException e) {
      e.printStackTrace();
    }
  }

  /** Handle a line command character. */
  private void lineCommandChar(String c) {
    int curpos = getCaretPosition();
    int row    = getRowNumber(curpos);
    int offs   = getOffsetInRow(curpos);
    setRowLineCommandChar(offs, c, row);
    int pos = getCaretPosition();
    replaceRange(c, pos, pos+1);
    setCaretPosition(wrappedCursorPos(++pos));
  }

  /** Delete char at current pos */
  private void deleteCharAtCursor () {
    int pos = getCaretPosition();
    int currowaddr = getRowOffsetFromCurpos(pos);
    int endcurrow  = toEndOfSrcdta(currowaddr);

    try {
      String right_text = getText(pos + 1, endcurrow - (pos + 1)) + " "; // add blank to left-shifting string
      replaceRange(right_text, pos, endcurrow);
      setRowChanged(true, getRowNumber(pos));
    }
    catch (javax.swing.text.BadLocationException e) {
      e.printStackTrace();
    }
    setCaretPosition(pos);
  }

  /** Delete line command char at current pos */
  private void deleteLineCommandCharAtCursor () {
    int pos    = getCaretPosition();
    int rownum = getRowNumber(pos);
    int offs   = getOffsetInRow(pos);
    int currowaddr  = getRowOffsetFromCurpos(pos);
    int endlinecmd  = toEndOfSrcseq(currowaddr);

    try {
      String right_text = getText(pos + 1, endlinecmd - (pos + 1)) + " "; // add blank to left-shifting string
      replaceRange(right_text, pos, endlinecmd);
    }
    catch (javax.swing.text.BadLocationException e) {
      e.printStackTrace();
    }
    deleteRowLineCommandChar(offs, rownum);
    setCaretPosition(pos);
  }

  /** Delete char at current pos */
  private void backspaceAtCursor () {
    int pos = getCaretPosition();
    int currowaddr = getRowOffsetFromCurpos(pos);
    int endcurrow  = toEndOfSrcdta(currowaddr);

    // Prevent illegal backspace into protected region or on protected row.
    if (inProtectedRegion(pos - 1) | currowaddr == pos) { // Backspace illegally
      beep();
    }
    else {
      try {
        replaceRange (getText(pos, endcurrow - pos) + " ", pos - 1, endcurrow);
        setRowChanged(true, getRowNumber(pos));
        setCaretPosition(pos - 1);
      }
      catch (javax.swing.text.BadLocationException e) {
        e.printStackTrace();
        beep();
      }
    }
  }

  /** Backspace over line command char at location before pos */
  private void backspaceLineCommandAtCursor () {
    int pos = getCaretPosition();
    int rownum = getRowNumber(pos);
    int offs   = getOffsetInRow(pos);
    int currowaddr  = getRowOffsetFromCurpos(pos);
    int endlinecmd  = toEndOfSrcseq(currowaddr);

    // illegal backspace head of line or protected line
    if (currowaddr == pos | onProtectedRow(pos)) {
      beep();
    }
    else { // S'okay.
      try {
        replaceRange (getText(pos, endlinecmd - pos) + " ", pos - 1, endlinecmd);
        backspaceRowLineCommandChar(offs, rownum);
        setCaretPosition(pos - 1);
      }
      catch (javax.swing.text.BadLocationException e) {
        e.printStackTrace();
        beep();
      }
    }
  }

  /** Do tab processing. !!! Under construction !!! */
  private void tabAtCursor() {
    int pos = getCaretPosition();
    setCaretPosition(rightCursorPos(pos,4));
  }

  /**************/
  /* Key Events */
  /**************/

  /** Tab around the display, possibly inserting */
  private void keyEventTabChar(java.awt.event.KeyEvent e) { // Arg not currently used, just for symmetry.
    int curpos = getCaretPosition();   // Current cursor pos
    if (inserting && inProtectedRegion(curpos)) { // Protected region checking
      beep();
    }
    else {
      if (e.getKeyCode() == java.awt.event.KeyEvent.VK_UNDEFINED) {
        tabAtCursor();
      }
    }
    e.consume();
  }

  /** Backspace deleting backwards and dragging line */
  private void keyEventBackspaceChar(java.awt.event.KeyEvent e) { // Arg not currently used, just for symmetry.
    int curpos = getCaretPosition();   // Current cursor pos
    if (e.getKeyCode() == java.awt.event.KeyEvent.VK_UNDEFINED) { // Any backspace combo just does backspace for now
      if (inLineCommandRegion(curpos - 1)) {
        backspaceLineCommandAtCursor();
      }
      else {
        backspaceAtCursor();
      }
      e.consume();
    }
  }

  /** ESC key means clear line attributes and redisplay */
  private void keyEventEscape(java.awt.event.KeyEvent e) {
    if (e.getKeyCode() == java.awt.event.KeyEvent.VK_UNDEFINED) {
      myController.eventClearRedisplay(e);
    }
    e.consume(); // In any case, consume event.
  }

  /** ENTER key means process changed, clear line attributes and redisplay */
  private void keyEventEnter(java.awt.event.KeyEvent e) {
    if (e.getKeyCode() == java.awt.event.KeyEvent.VK_UNDEFINED) {
      myController.eventProcessChanged(e);
    }
    e.consume(); // In any case, consume event.
  }

  /*********************/
  /* Key Event Process */
  /*********************/

  /** Process certain key events before textedit gets 'em */
  protected void processKeyEvent(java.awt.event.KeyEvent e) {
    // int keyCode = e.getKeyCode();
    int keyChar = e.getKeyChar();
    // int keyModifiers = e.getModifiers();
    // String keyModifiersText = e.getKeyModifiersText(keyModifiers);
    // String keyText = e.getKeyText(keyCode);
    // boolean actionKey = e.isActionKey();
    // String paramString = e.paramString();

    /* Debug */
    /*
     * System.out.println
     * ("Keycode " + keyCode + " Keychar " + keyChar + " Modifiers " + keyModifiers
     * + " Modtext " + keyModifiersText + " Text " + keyText + " Actionkey " + actionKey
     * + " Parmstring " + paramString);
     */

    if (keyChar == '\t') {
      keyEventTabChar(e);
    }
    else if (keyChar == '\b') {
      keyEventBackspaceChar(e);
    }
    else if (keyChar == '\033') {
      keyEventEscape(e);
    }
    else if (keyChar == '\012') {
      keyEventEnter(e);
    }
    else {
      super.processKeyEvent(e);
    }
  }

  /***********/
  /* Actions */
  /***********/

  public void beep() {
    getToolkit().beep();
  }

  /** Action performed for most visible characters (all?) */
  private void actionRegularChar(java.awt.event.ActionEvent e) {
    String c   = e.getActionCommand(); // Character typed
    // System.out.println(e.toString());
    short charval = (short) c.charAt(0);
    // System.out.println(charval);
    if (127 != charval) { // This catches ctrl-backspace which doesn't seem to be easily interceptable otherwise.
      int curpos = getCaretPosition();   // Current cursor pos
      if ((e.getModifiers()
      & (java.awt.event.ActionEvent.ALT_MASK
      | java.awt.event.ActionEvent.CTRL_MASK
      | java.awt.event.ActionEvent.META_MASK
      ))
      == 0
      ) { // If it wasn't an alt meta or control character
        if (inProtectedRegion(curpos)) { // Protected region checking
          if (c.equals(" ")) { // If it's a space in protected, use it as move right.
            moveCursorRight();
          }
          else {
            beep();
          }
        }
        else if (inLineCommandRegion(curpos)) {
          lineCommandChar(c);
        }
        else if (inserting) {
          insertChar(c);
        }
        else {
          typeoverChar(c);
        }
      }
    }
    else { // ctrl-backspace has snuck through into actionRegularChar()
      beep();
    }
  }

  /** Delete pulling back rest of line, where alteration allowed */
  private void actionDeleteChar(java.awt.event.ActionEvent e) { // Arg not currently used, just for symmetry.
    int curpos = getCaretPosition();   // Current cursor pos

    if (inProtectedRegion(curpos)) { // Protected region checking
      beep();
    }
    else if (inLineCommandRegion(curpos)) {
      deleteLineCommandCharAtCursor();
    }
    else  {
      deleteCharAtCursor();
    }
  }

  /**********/
  /* Keymap */
  /**********/

  /** Some special keys are handled */
  private void addCustomKeymap() {
    // keymap = addKeymap("com.softwoehr.jtopencontrib.meu.keymap", getKeymap());
    keymap = addKeymap("com.softwoehr.jtopencontrib.meu.keymap", null);

    /* Actions which can be taken locally */

    keymap.setDefaultAction( new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        actionRegularChar(e);
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_DELETE, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        actionDeleteChar(e);
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_HOME, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorHome();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_END, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorEnd();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_HOME, java.awt.event.KeyEvent.CTRL_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorTopHome();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_END, java.awt.event.KeyEvent.CTRL_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorBottomEnd();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_HOME, java.awt.event.KeyEvent.SHIFT_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorLineArea();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_END, java.awt.event.KeyEvent.SHIFT_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorSrcdtaEnd();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_UP, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorUp();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_DOWN, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorDown();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_LEFT, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorLeft();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_RIGHT, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorRight();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_PAGE_UP, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorPageUp();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_PAGE_DOWN, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        moveCursorPageDown();
      }
    });

    //////////////////////////////////////////////////////////////
    /* Actions which require the intervention of the controller */
    //////////////////////////////////////////////////////////////

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_INSERT, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        myController.actionToggleInserting(e);
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_PAGE_UP, java.awt.event.KeyEvent.CTRL_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        myController.actionScrollBack(e);
        wrapCursor();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_PAGE_DOWN, java.awt.event.KeyEvent.CTRL_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        myController.actionScrollAhead(e);
        wrapCursor();
      }
    });

    ///////////
    // FKeys //
    ///////////

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_F3, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        myController.actionClose(e);
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_F4, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        myController.actionSave(e);
        wrapCursor();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_F5, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        myController.actionSplitJoin(e);
        wrapCursor();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_F10, 0) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        myController.actionToggleFocus(e);
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_V, java.awt.event.KeyEvent.CTRL_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        System.out.println("Ctrl-V");
        wrapCursor();
      }
    });

    keymap.addActionForKeyStroke(javax.swing.KeyStroke.getKeyStroke
    (java.awt.event.KeyEvent.VK_X, java.awt.event.KeyEvent.CTRL_MASK) ,
    new javax.swing.AbstractAction () {
      public void actionPerformed(java.awt.event.ActionEvent e) {
        System.out.println("Ctrl-X");
        wrapCursor();
      }
    });
    setKeymap(keymap);
  }
  /**********************/
  /* Application Syntax */
  /**********************/

  /** Get an application-wide property */
  private int getIntProperty(String p)
  throws MEUProperties.NoSuchPropertyException,
  MEUProperties.InvalidIntPropertyException
  {
    return myController.getMEU().getIntProperty(p);
  }

  /** Get an application-wide property */
  private String getProperty(String p) {
    return myController.getMEU().getProperty(p);
  }

  private java.awt.Color toColor(String s) {

    java.awt.Color c = java.awt.Color.yellow;

    if (s.equals("black")) {
      c = java.awt.Color.black;
    }
    else if (s.equals("blue")) {
      c = java.awt.Color.blue;
    }
    else if (s.equals("cyan")) {
      c = java.awt.Color.cyan;
    }
    else if (s.equals("darkGray")) {
      c = java.awt.Color.darkGray;
    }
    else if (s.equals("gray")) {
      c = java.awt.Color.gray;
    }
    else if (s.equals("green")) {
      c = java.awt.Color.green;
    }
    else if (s.equals("lightGray")) {
      c = java.awt.Color.lightGray;
    }
    else if (s.equals("magenta")) {
      c = java.awt.Color.magenta;
    }
    else if (s.equals("orange")) {
      c = java.awt.Color.orange;
    }
    else if (s.equals("pink")) {
      c = java.awt.Color.pink;
    }
    else if (s.equals("red")) {
      c = java.awt.Color.red;
    }
    else if (s.equals("white")) {
      c = java.awt.Color.white;
    }
    return c;
  }
}
